from policyengine_us.model_api import *


class is_male(Variable):
    value_type = bool
    entity = Person
    label = "is male"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        return ~person("is_female", period)



from policyengine_us.model_api import *


class salt_refund_last_year(Variable):
    value_type = float
    entity = TaxUnit
    label = "SALT refund last year"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class va_spouse_tax_adjustment_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for Virginia's spouse tax adjustment"
    defined_for = StateCode.VA
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=19"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        agi_less_exemptions_person = person(
            "va_agi_less_exemptions_person", period
        )
        agi_less_exemptions = where(
            head_or_spouse, agi_less_exemptions_person, np.inf
        )
        smaller_exemptions = tax_unit.min(agi_less_exemptions)
        joint = tax_unit("tax_unit_is_joint", period)
        return (smaller_exemptions > 0) & joint



from policyengine_us.model_api import *


class pell_grant_min_fpg_percent_limit(Variable):
    value_type = float
    unit = "/1"
    entity = Person
    definition_period = YEAR
    label = "The maximum FPG percent to qualify for the minimum Pell Grant"

    def formula(person, period, parameters):
        household_type = person("pell_grant_household_type", period)
        limits = parameters(
            period
        ).gov.ed.pell_grant.sai.fpg_fraction.min_pell_limits
        not_parent = person.tax_unit("tax_unit_child_dependents", period) == 0

        independent = (
            household_type == household_type.possible_values.INDEPENDENT_SINGLE
        ) | (
            household_type
            == household_type.possible_values.INDEPENDENT_NOT_SINGLE
        )

        independent_parent = independent & not_parent

        return where(
            independent_parent,
            limits.INDEPENDENT_NOT_PARENT,
            limits[household_type],
        )



from policyengine_us.model_api import *


class az_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona Itemized Deductions"
    unit = USD
    documentation = "Arizona Form 140 Schedule A"
    definition_period = YEAR
    defined_for = StateCode.AZ
    reference = (
        "https://law.justia.com/codes/arizona/2022/title-43/section-43-1042/",
        "https://azdor.gov/forms/individual/itemized-deduction-adjustments-form",
        "https://azdor.gov/forms/individual/form-140-resident-personal-income-tax-form-calculating",
    )
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        deductions = [
            deduction
            for deduction in p.itemized_deductions
            if deduction
            not in [
                "medical_expense_deduction",
                "charitable_deduction",
            ]
        ]
        federal_deductions = add(tax_unit, period, deductions)
        # Arizona allows a complete deduction for medical and dental expenses
        medical_expenses = add(
            tax_unit, period, ["medical_out_of_pocket_expenses"]
        )

        # Adjustments to Charitable Contributions
        # Amount of charitable contributions for which you are claiming
        # a credit under Arizona law.
        charitable_deduction = tax_unit("charitable_deduction", period)
        charitable_contributions_credit = tax_unit(
            "az_charitable_contributions_credit", period
        )
        # The charitable deduction is reduced by the amount which is used for the
        # Arizona charitable contributions credit,
        # assuming that the same charitable contributions
        charitable_deduction_after_credit = max_(
            charitable_deduction - charitable_contributions_credit, 0
        )
        # The state and local income tax is reduced by the amount which is used
        # to claim the Arizona charitable contributions credit
        # Since the Arizona charitable contributions credit is based on contributions
        # to qulalifying foster care organizations, we do not reduce the salt deduction
        return (
            federal_deductions
            + medical_expenses
            + charitable_deduction_after_credit
        )



from policyengine_us.model_api import *


class ca_foster_youth_tax_credit_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the California foster youth tax credit"
    definition_period = YEAR
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=RTC&sectionNum=17052.2."
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        was_in_foster_care = person("was_in_foster_care", period)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return was_in_foster_care & is_head_or_spouse



from policyengine_us.model_api import *


class il_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ilga.gov/legislation/fulltext.asp?DocName=&SessionId=112&GA=103&DocTypeId=HB&DocNum=4917&GAID=17&LegID=152789&SpecSess=&Session="
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.il.tax.income.credits.ctc
        person = tax_unit.members
        age = person("age", period)
        age_eligible_child = age < p.age_limit
        federal_ctc_eligible_child = person("ctc_qualifying_child", period)
        eligible_child = age_eligible_child & federal_ctc_eligible_child
        eligible_child_present = tax_unit.any(eligible_child)
        state_eitc = tax_unit("il_eitc", period)
        return eligible_child_present * state_eitc * p.rate



from policyengine_us.model_api import *


class ia_net_income(Variable):
    value_type = float
    entity = Person
    label = "Iowa net income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        gross_income = person("ia_gross_income", period)
        income_adjustments = person("ia_income_adjustments", period)
        net_income = gross_income - income_adjustments
        # allocate any dependent net_income to tax unit head
        is_dependent = person("is_tax_unit_dependent", period)
        sum_dep_net_income = person.tax_unit.sum(is_dependent * net_income)
        is_head = person("is_tax_unit_head", period)
        return ~is_dependent * net_income + is_head * sum_dep_net_income



from policyengine_us.model_api import *


class SNAPUttilityAllowanceType(Enum):
    SUA = "Standard Utility Allowance"
    LUA = "Limited Utility Allowance"
    IUA = "Individual Utility Allowance"
    NONE = "None"


class snap_utility_allowance_type(Variable):
    value_type = Enum
    possible_values = SNAPUttilityAllowanceType
    entity = SPMUnit
    label = "SNAP utility allowance eligibility"
    default_value = SNAPUttilityAllowanceType.NONE
    documentation = (
        "The type of utility allowance that is eligible for the SPM unit"
    )
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        distinct_utility_bills = spm_unit(
            "count_distinct_utility_expenses", period
        )
        lua = parameters(
            period
        ).gov.usda.snap.income.deductions.utility.limited
        region = spm_unit.household("snap_utility_region_str", period)
        always_sua = spm_unit(
            "snap_state_using_standard_utility_allowance", period
        )
        has_heating_cooling = spm_unit("has_heating_cooling_expense", period)
        lua_is_defined = lua.active[region].astype(bool)
        return select(
            [
                has_heating_cooling | always_sua,
                lua_is_defined & (distinct_utility_bills >= 2),
                distinct_utility_bills > 0,
            ],
            [
                SNAPUttilityAllowanceType.SUA,
                SNAPUttilityAllowanceType.LUA,
                SNAPUttilityAllowanceType.IUA,
            ],
            default=SNAPUttilityAllowanceType.NONE,
        )



from policyengine_us.model_api import *


class basic_income_before_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "Basic income before phase-outs"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.contrib.ubi_center.basic_income
        # Start with flat person-level amount.
        total_flat_amount = p.amount.person.flat * tax_unit(
            "tax_unit_size", period
        )
        # Add per-age person-level amount.
        person = tax_unit.members
        age = person("age", period)
        amount_by_age = p.amount.person.by_age.calc(age)
        total_amount_by_age = tax_unit.sum(amount_by_age)
        # If available, apply a marriage bonus
        married = tax_unit.family("is_married", period)
        marriage_bonus_rate = (
            p.amount.person.marriage_bonus * total_amount_by_age
        )
        post_marriage_bonus_amount = total_amount_by_age + marriage_bonus_rate
        applicable_amount_by_age = where(
            married, post_marriage_bonus_amount, total_amount_by_age
        )
        # Now compute FPG amount.
        fpg = tax_unit("tax_unit_fpg", period)
        fpg_amount = p.amount.tax_unit.fpg_percent * fpg

        # Disability amount
        disabled = person("is_ssi_disabled", period)
        disabled_amount = tax_unit.sum(disabled * p.amount.person.disability)
        base_amount = (
            total_flat_amount
            + applicable_amount_by_age
            + fpg_amount
            + disabled_amount
        )
        if p.phase_in.in_effect:
            phase_in_amount = tax_unit("basic_income_phase_in", period)
            return min_(base_amount, phase_in_amount)
        return base_amount



from policyengine_us.model_api import *


class nj_total_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey total deductions to income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NJ

    adds = ["nj_medical_expense_deduction"]



from policyengine_us.model_api import *


class taxable_ss_magi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Modified adjusted gross income (SS)"
    unit = USD
    documentation = "Income used to determine taxability of Social Security."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/86"

    def formula(tax_unit, period, parameters):
        irs = parameters(period).gov.irs
        gross_income_sources = irs.gross_income.sources
        ss_magi = irs.social_security.taxability.income
        income_sources_without_ss = [
            income_source
            for income_source in gross_income_sources
            if income_source
            not in [
                "taxable_social_security",
                "taxable_unemployment_compensation",
            ]
        ]
        if "taxable_unemployment_compensation" in gross_income_sources:
            # Reforms which drop the UI variable from gross income should
            # trigger SS-related MAGI to drop it too.
            income_sources_without_ss.append("unemployment_compensation")
        income_sources_without_ss.append("tax_exempt_interest_income")
        gross_income = 0
        person = tax_unit.members
        not_dependent = ~person("is_tax_unit_dependent", period)
        for source in income_sources_without_ss:
            # Add positive values only - losses are deducted later.
            gross_income += not_dependent * max_(
                0, add(person, period, [source])
            )
        gross_income = tax_unit.sum(gross_income)
        above_the_line_deductions = irs.ald.deductions
        revoked_deductions = ss_magi.revoked_deductions
        deductions = [
            deduction
            for deduction in above_the_line_deductions
            if deduction not in revoked_deductions
        ]
        total_deductions = add(tax_unit, period, deductions)
        return max_(0, gross_income - total_deductions)



from policyengine_us.model_api import *


class wagering_losses_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wagering losses deduction"
    unit = USD
    documentation = "Deduction from taxable income for gambling losses, capped at gambling winnings."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/165#d"

    def formula(tax_unit, period, parameters):
        gambling_losses = add(tax_unit, period, ["gambling_losses"])
        gambling_winnings = add(tax_unit, period, ["gambling_winnings"])
        return min_(gambling_losses, gambling_winnings)



from policyengine_us.model_api import *


class nm_low_income_comprehensive_tax_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico low income comprehensive tax rebate"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=58",
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=70",
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/856ebf4b-3814-49dd-8631-ebe579d6a42b/Personal%20Income%20Tax.pdf#page=67",
        "https://casetext.com/statute/new-mexico-statutes-1978/chapter-7-taxation/article-2-income-tax-general-provisions/section-7-2-14-low-income-comprehensive-tax-rebate?sort=relevance&type=regulation&tab=keyword&jxs=&resultsNav=false",
    )
    # 7-2-7.6. 2021 INCOME TAX REBATE
    # SECTION II: LOW INCOME COMPREHENSIVE TAX REBATE
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nm.tax.income.rebates.low_income

        agi = tax_unit("nm_modified_gross_income", period)
        exemptions = tax_unit(
            "nm_low_income_comprehensive_tax_rebate_exemptions", period
        )

        rebate = select(
            [
                exemptions == 1,
                exemptions == 2,
                exemptions == 3,
                exemptions == 4,
                exemptions == 5,
            ],
            [
                p.amount.one_exemption.calc(agi, right=True),
                p.amount.two_exemptions.calc(agi, right=True),
                p.amount.three_exemptions.calc(agi, right=True),
                p.amount.four_exemptions.calc(agi, right=True),
                p.amount.five_exemptions.calc(agi, right=True),
            ],
            default=p.amount.six_exemptions.calc(agi, right=True),
        )

        filing_status = tax_unit("filing_status", period)
        divisor = where(
            filing_status == filing_status.possible_values.SEPARATE,
            p.divisor,
            1,
        )
        return rebate / divisor



from policyengine_us.model_api import *


class tx_tanf_countable_resources(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas TANF countable resources"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1210-general-policy",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1220-limits",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Countable resources = liquid resources + excess vehicle value
        # Household must have countable resources ≤ $1,000 to be eligible

        # Liquid resources (cash, checking, savings) - stocks use period.this_year
        cash_assets = spm_unit("spm_unit_cash_assets", period.this_year)

        # Vehicle resources (apply vehicle exemption: one automobile up to $4,650)
        vehicle_value = spm_unit.household(
            "household_vehicles_value", period.this_year
        )
        p = parameters(period).gov.states.tx.tanf.resources
        countable_vehicle_value = max_(vehicle_value - p.vehicle_exemption, 0)

        # Total countable resources (no period conversion - resources are stocks)
        return cash_assets + countable_vehicle_value



from policyengine_us.model_api import *


class income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "Federal income tax before refundable credits"
    documentation = "Income tax liability (including other taxes) after non-refundable credits are used, but before refundable credits are applied"

    def formula(tax_unit, period, parameters):
        if parameters(
            period
        ).gov.contrib.ubi_center.flat_tax.abolish_federal_income_tax:
            return 0
        else:
            added_components = add(
                tax_unit,
                period,
                [
                    "income_tax_before_credits",
                    "net_investment_income_tax",
                    "recapture_of_investment_credit",
                    "unreported_payroll_tax",
                    "qualified_retirement_penalty",
                ],
            )
            subtracted_components = add(
                tax_unit, period, ["income_tax_capped_non_refundable_credits"]
            )
            return added_components - subtracted_components



from policyengine_us.model_api import *


class mt_income_tax_before_non_refundable_credits_joint(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana income tax before refundable credits when married couples file jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = ["mt_capital_gains_tax_joint", "mt_regular_income_tax_joint"]



from policyengine_us.model_api import *
from numpy import ceil


class ctc_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC reduction from income"
    unit = USD
    documentation = "Reduction of the total CTC due to income."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # TCJA's phase-out changes are purely parametric so don't require
        # structural reform.

        # The ARPA CTC has two phase-outs: the original, and a new phase-out
        # applying before and only to the increase in the maximum CTC under ARPA.

        # Start with the normal phase-out.
        income = tax_unit("adjusted_gross_income", period)
        p = parameters(period).gov.irs.credits.ctc.phase_out
        phase_out_threshold = tax_unit("ctc_phase_out_threshold", period)
        excess = max_(0, income - phase_out_threshold)
        increments = ceil(excess / p.increment)
        return increments * p.amount



from policyengine_us.model_api import *


class ca_riv_liheap_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Riverside County Low Income Home Energy Assistance Program (LIHEAP) countable income"
    definition_period = YEAR
    defined_for = "in_riv"
    reference = "https://capriverside.org/sites/g/files/aldnop136/files/2024-12/2025%20LIHEAP%20CAP%20APPLICATION%20ENGLISH.pdf#page=3"

    adds = "gov.local.ca.riv.cap.liheap.countable_income.sources"



from policyengine_us.model_api import *


class dependent_care_employer_benefits(Variable):
    value_type = float
    entity = Person
    label = "Dependent care benefits received from an employer"
    # description = "Amounts you received as an employee should be shown in Box 10 of your federal Form(s) W-2. If you were self-employed or a partner, include amounts you received under a dependent care assistance program from your sole proprietorship or partnership"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class mt_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana income tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        filing_separately = tax_unit("mt_files_separately", period)
        itax_indiv = add(tax_unit, period, ["mt_income_tax_indiv"])
        itax_joint = tax_unit("mt_income_tax_joint", period)
        return where(filing_separately, itax_indiv, itax_joint)



from policyengine_us.model_api import *


class medicaid_income_level(Variable):
    value_type = float
    entity = Person
    label = "Medicaid/CHIP-related income level"
    unit = "/1"
    documentation = (
        "Modified AGI as a fraction of current-year federal poverty line."
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        return person.tax_unit("tax_unit_medicaid_income_level", period)



from policyengine_us.model_api import *


class premium_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Affordable Care Act Premium Tax Credit"
    unit = USD
    definition_period = MONTH

    adds = ["aca_ptc"]



from policyengine_us.model_api import *


def create_boost_middle_class_tax_credit() -> Reform:
    class boost_middle_class_tax_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "BOOST Middle Class Tax Credit"
        unit = USD
        definition_period = YEAR
        reference = "https://d12t4t5x3vyizu.cloudfront.net/tlaib.house.indigov.us/uploads/2022/05/TLAIB_046_xml.pdf"

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.harris.lift.middle_class_tax_credit
            filing_status = tax_unit("filing_status", period)
            joint = filing_status == filing_status.possible_values.JOINT
            amount = where(joint, p.cap * p.joint_multiplier, p.cap)
            agi = tax_unit("adjusted_gross_income", period)

            excess = max_(0, agi - p.phase_out.start[filing_status])
            phase_out_rate = min_(1, excess / p.phase_out.width[filing_status])
            reduction = amount * phase_out_rate
            age = tax_unit.members("age", period)
            age_eligible = age >= p.age_threshold
            eligible = tax_unit.any(age_eligible)
            return max_(0, amount - reduction) * eligible

    class income_tax_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "federal refundable income tax credits"
        unit = USD

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.irs.credits
            previous_credits = add(tax_unit, period, p.refundable)
            middle_class_credit = tax_unit(
                "boost_middle_class_tax_credit", period
            )
            p_boost = parameters(
                period
            ).gov.contrib.congress.tlaib.boost.middle_class_tax_credit
            if p_boost.administered_through_ssa:
                return previous_credits
            return middle_class_credit + previous_credits

    class household_benefits(Variable):
        value_type = float
        entity = Household
        label = "benefits"
        unit = USD
        definition_period = YEAR

        def formula(household, period, parameters):
            BENEFITS = [
                "social_security",
                "ssi",
                "snap",
                "wic",
                "free_school_meals",
                "reduced_price_school_meals",
                "spm_unit_broadband_subsidy",
                "tanf",
                "high_efficiency_electric_home_rebate",
                "residential_efficiency_electrification_rebate",
                "unemployment_compensation",
                # Contributed.
                "basic_income",
                "spm_unit_capped_housing_subsidy",
                "household_state_benefits",
            ]
            previous_benefits = add(household, period, BENEFITS)
            middle_class_credit = add(
                household, period, ["boost_middle_class_tax_credit"]
            )
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.boost.middle_class_tax_credit
            if p.administered_through_ssa is False:
                return previous_benefits
            return middle_class_credit + previous_benefits

    class spm_unit_benefits(Variable):
        value_type = float
        entity = SPMUnit
        label = "Benefits"
        definition_period = YEAR
        unit = USD

        def formula(spm_unit, period, parameters):
            BENEFITS = [
                "social_security",
                "ssi",
                "ma_state_supplement",  # Massachusetts benefits
                # California programs.
                "ca_cvrp",  # California Clean Vehicle Rebate Project.
                # Colorado programs.
                "co_ccap_subsidy",
                "co_state_supplement",
                "co_oap",
                "snap",
                "wic",
                "free_school_meals",
                "reduced_price_school_meals",
                "spm_unit_broadband_subsidy",
                "spm_unit_energy_subsidy",
                "tanf",
                "high_efficiency_electric_home_rebate",
                "residential_efficiency_electrification_rebate",
                "unemployment_compensation",
                # Contributed.
                "basic_income",
                "ny_drive_clean_rebate",
            ]
            if parameters(period).gov.contrib.ubi_center.flat_tax.deduct_ptc:
                BENEFITS.append("aca_ptc")
            if not parameters(period).gov.hud.abolition:
                BENEFITS.append("spm_unit_capped_housing_subsidy")
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.boost.middle_class_tax_credit
            middle_class_tax_credit = add(
                spm_unit, period, ["boost_middle_class_tax_credit"]
            )
            previous_benefits = add(spm_unit, period, BENEFITS)
            if p.administered_through_ssa is False:
                return previous_benefits
            return add(spm_unit, period, BENEFITS) + middle_class_tax_credit

    class reform(Reform):
        def apply(self):
            self.update_variable(boost_middle_class_tax_credit)
            self.update_variable(income_tax_refundable_credits)
            self.update_variable(household_benefits)
            self.update_variable(spm_unit_benefits)

    return reform


def create_boost_middle_class_tax_credit_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_boost_middle_class_tax_credit()

    p = parameters(period).gov.contrib.harris.lift.middle_class_tax_credit

    if p.in_effect:
        return create_boost_middle_class_tax_credit()
    else:
        return None


boost_middle_class_tax_credit = create_boost_middle_class_tax_credit_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class spm_unit_weight(Variable):
    value_type = float
    entity = SPMUnit
    label = "SPM unit weight"
    definition_period = YEAR
    uprating = "calibration.gov.census.populations.total"

    def formula(spm_unit, period, parameters):
        # Use household weights if not provided
        return spm_unit.household("household_weight", period)



from policyengine_us.model_api import *


class income_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "Federal income tax"
    documentation = "Total federal individual income tax liability."

    def formula(person, period, parameters):
        if parameters(
            period
        ).gov.contrib.ubi_center.flat_tax.abolish_federal_income_tax:
            return 0
        else:
            added_components = add(
                person, period, ["income_tax_before_refundable_credits"]
            )
            subtracted_components = add(
                person, period, ["income_tax_refundable_credits"]
            )
            return added_components - subtracted_components



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_income_security_package() -> Reform:
    class baby_bonus_act_payment(Variable):
        value_type = float
        entity = Person
        definition_period = YEAR
        label = "Baby Bonus Act payment"
        unit = USD
        documentation = (
            "One-time payment for newborns under the Baby Bonus Act"
        )
        reference = "placeholder - bill not yet introduced"

        def formula(person, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.baby_bonus_act

            # Baby bonus is for those born in 2026 and after, paid in their birth year
            birth_year = person("birth_year", period)

            age = person("age", period)

            # Payment only in the birth year for those born 2026+
            is_eligible = (birth_year >= p.min_birth_year) & (
                age < p.age_limit
            )

            return is_eligible * p.amount

    class boost_act_payment(Variable):
        value_type = float
        entity = Person
        definition_period = YEAR
        label = "BOOST Act payment"
        unit = USD
        documentation = (
            "Monthly payments under the BOOST Act for eligible adults"
        )
        reference = "placeholder - bill not yet introduced"

        def formula(person, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.boost_act

            age = person("age", period)
            is_eligible = (age >= p.min_age) & (age <= p.max_age)

            annual_amount = p.amount * MONTHS_IN_YEAR

            return is_eligible * annual_amount

    class boost_act_tax(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "BOOST Act supplemental tax"
        unit = USD
        documentation = "Supplemental tax on AGI to fund the BOOST Act"
        reference = "placeholder - bill not yet introduced"

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.boost_act

            agi = tax_unit("adjusted_gross_income", period)
            filing_status = tax_unit("filing_status", period)

            excess_agi = max_(agi - p.tax.threshold[filing_status], 0)

            return excess_agi * p.tax.rate

    class ecpa_child_benefit(Variable):
        value_type = float
        entity = Person
        definition_period = YEAR
        label = "End Child Poverty Act child benefit"
        unit = USD
        documentation = (
            "Universal child benefit under the End Child Poverty Act"
        )
        reference = "placeholder - bill not yet introduced"

        def formula(person, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.end_child_poverty_act

            age = person("age", period)
            is_eligible = age < p.child_benefit.age_limit

            # Use FPG additional_person amount (CONTIGUOUS_US for all states), rounded to nearest $10
            p_fpg = parameters(period).gov.hhs.fpg
            fpg_amount = p_fpg.additional_person.CONTIGUOUS_US
            amount = round(fpg_amount / 10) * 10

            return is_eligible * amount

    class ecpa_adult_dependent_credit(Variable):
        value_type = float
        entity = Person
        definition_period = YEAR
        label = "End Child Poverty Act adult dependent credit"
        unit = USD
        documentation = (
            "Credit for adult dependents under the End Child Poverty Act"
        )
        reference = "placeholder - bill not yet introduced"

        def formula(person, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.end_child_poverty_act

            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)

            is_eligible = (
                age >= p.adult_dependent_credit.min_age
            ) & is_dependent

            return is_eligible * p.adult_dependent_credit.amount

    class ecpa_filer_credit(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "End Child Poverty Act filer credit"
        unit = USD
        documentation = "Filer credit under the End Child Poverty Act for eligible tax filers"
        reference = "placeholder - bill not yet introduced"

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.end_child_poverty_act

            person = tax_unit.members
            age = person("age", period)

            is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

            is_eligible_age = (age >= p.filer_credit.eligibility.min_age) & (
                age < p.filer_credit.eligibility.max_age
            )
            is_eligible_filer = is_head_or_spouse & is_eligible_age

            has_eligible_filer = tax_unit.any(is_eligible_filer)

            filing_status = tax_unit("filing_status", period)
            amount = p.filer_credit.amount[filing_status]

            agi = tax_unit("adjusted_gross_income", period)
            excess_agi = max_(
                agi - p.filer_credit.phase_out.start[filing_status], 0
            )
            phase_out = excess_agi * p.filer_credit.phase_out.rate

            phased_out_amount = max_(amount - phase_out, 0)

            return has_eligible_filer * phased_out_amount

    class household_benefits(Variable):
        value_type = float
        entity = Household
        label = "benefits"
        unit = USD
        definition_period = YEAR

        def formula(household, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package

            BENEFITS = [
                "social_security",
                "ssi",
                "snap",
                "wic",
                "free_school_meals",
                "reduced_price_school_meals",
                "spm_unit_broadband_subsidy",
                "tanf",
                "high_efficiency_electric_home_rebate",
                "residential_efficiency_electrification_rebate",
                "unemployment_compensation",
                # Contributed.
                "basic_income",
                "spm_unit_capped_housing_subsidy",
                "household_state_benefits",
            ]

            # Add Income Security Package benefits if active
            if p.baby_bonus_act.in_effect:
                BENEFITS.append("baby_bonus_act_payment")
            if p.boost_act.in_effect:
                BENEFITS.append("boost_act_payment")
            if p.end_child_poverty_act.in_effect:
                BENEFITS.append("ecpa_child_benefit")

            return add(household, period, BENEFITS)

    class spm_unit_benefits(Variable):
        value_type = float
        entity = SPMUnit
        label = "Benefits"
        definition_period = YEAR
        unit = USD

        def formula(spm_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package

            BENEFITS = [
                "social_security",
                "ssi",
                "ma_state_supplement",  # Massachusetts benefits
                # California programs.
                "ca_cvrp",  # California Clean Vehicle Rebate Project.
                # Colorado programs.
                "co_ccap_subsidy",
                "co_state_supplement",
                "co_oap",
                "snap",
                "wic",
                "free_school_meals",
                "reduced_price_school_meals",
                "spm_unit_broadband_subsidy",
                "spm_unit_energy_subsidy",
                "tanf",
                "high_efficiency_electric_home_rebate",
                "residential_efficiency_electrification_rebate",
                "unemployment_compensation",
                # Contributed.
                "basic_income",
                "ny_drive_clean_rebate",
            ]

            # Add Income Security Package benefits if active
            if p.baby_bonus_act.in_effect:
                BENEFITS.append("baby_bonus_act_payment")
            if p.boost_act.in_effect:
                BENEFITS.append("boost_act_payment")
            if p.end_child_poverty_act.in_effect:
                BENEFITS.append("ecpa_child_benefit")

            if parameters(period).gov.contrib.ubi_center.flat_tax.deduct_ptc:
                BENEFITS.append("aca_ptc")
            if not parameters(period).gov.hud.abolition:
                BENEFITS.append("spm_unit_capped_housing_subsidy")

            return add(spm_unit, period, BENEFITS)

    class income_tax(Variable):
        value_type = float
        entity = TaxUnit
        label = "federal income tax"
        unit = USD
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package

            # Start with income tax after non-refundable credits
            base_tax = tax_unit("income_tax_before_refundable_credits", period)

            # Subtract refundable credits
            refundable_credits = tax_unit(
                "income_tax_refundable_credits", period
            )

            reduced_tax = base_tax - refundable_credits

            # Add BOOST tax only if active
            if p.boost_act.in_effect:
                boost_tax = tax_unit("boost_act_tax", period)
                return reduced_tax + boost_tax
            else:
                return reduced_tax

    class income_tax_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "federal refundable income tax credits"
        unit = USD

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.end_child_poverty_act

            # Get standard refundable credits, excluding EITC and refundable CTC if ECPA is active
            standard_credits = parameters(period).gov.irs.credits.refundable

            if p.in_effect:
                # ECPA replaces EITC and refundable CTC
                CREDITS = [
                    c
                    for c in standard_credits
                    if c not in ["eitc", "refundable_ctc"]
                ]
            else:
                # Use all standard credits
                CREDITS = list(standard_credits)

            base_credits = add(tax_unit, period, CREDITS) if CREDITS else 0

            # Add ECPA credits only if active
            if p.in_effect:
                ecpa_filer = tax_unit("ecpa_filer_credit", period)
                ecpa_adult_dep = add(
                    tax_unit, period, ["ecpa_adult_dependent_credit"]
                )
                return base_credits + ecpa_filer + ecpa_adult_dep
            else:
                return base_credits

    class income_tax_non_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "federal non-refundable income tax credits"
        unit = USD

        def formula(tax_unit, period, parameters):
            # Get the base list of non-refundable credits from parameters
            p = parameters(period).gov.irs.credits
            base_credits = list(p.non_refundable)

            # Check if ECPA is active
            ecpa_params = parameters(
                period
            ).gov.contrib.congress.tlaib.income_security_package.end_child_poverty_act
            ecpa_active = ecpa_params.in_effect

            # If ECPA is active, remove non_refundable_ctc from the list
            if ecpa_active:
                CREDITS = [
                    c for c in base_credits if c != "non_refundable_ctc"
                ]
            else:
                CREDITS = base_credits

            return add(tax_unit, period, CREDITS)

    class reform(Reform):
        def apply(self):
            # Update all variables for the Income Security Package
            self.update_variable(baby_bonus_act_payment)
            self.update_variable(boost_act_payment)
            self.update_variable(boost_act_tax)
            self.update_variable(ecpa_child_benefit)
            self.update_variable(ecpa_adult_dependent_credit)
            self.update_variable(ecpa_filer_credit)
            self.update_variable(household_benefits)
            self.update_variable(spm_unit_benefits)
            self.update_variable(income_tax)
            self.update_variable(income_tax_refundable_credits)
            self.update_variable(income_tax_non_refundable_credits)

    return reform


def create_income_security_package_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_income_security_package()

    p = parameters.gov.contrib.congress.tlaib.income_security_package

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if (
            p(current_period).baby_bonus_act.in_effect
            or p(current_period).boost_act.in_effect
            or p(current_period).end_child_poverty_act.in_effect
        ):
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_income_security_package()
    else:
        return None


income_security_package = create_income_security_package_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class is_citizen_or_legal_immigrant(Variable):
    value_type = bool
    entity = Person
    label = "Is citizen or qualified noncitizen under federal law"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/8/1641"

    def formula(person, period, parameters):
        p = parameters(period).gov.dhs.immigration
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        is_citizen = (
            immigration_status == immigration_status.possible_values.CITIZEN
        )
        qualified_noncitizen = np.isin(
            immigration_status_str,
            p.qualified_noncitizen_status,
        )
        return qualified_noncitizen | is_citizen



from policyengine_us.model_api import *


class state_income_tax_reported(Variable):
    value_type = float
    entity = Person
    label = "reported State income tax"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class or_federal_tax_liability_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR federal tax liability subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/publication-or-17_101-431_2021.pdf#page=71",
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # Subsection 316.800
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        # calculate Oregon concept of federal income tax
        federal_itax = tax_unit("income_tax", period)
        federal_eitc = tax_unit("eitc", period)
        or_federal_income_tax = max_(0, federal_itax + federal_eitc)
        # limit subtraction based on caps scaled to federal AGI
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        p = (
            parameters(period)
            .gov.states["or"]
            .tax.income.subtractions.federal_tax_liability.cap
        )
        federal_agi = tax_unit("adjusted_gross_income", period)
        cap = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.SEPARATE,
                filing_status == status.SURVIVING_SPOUSE,
            ],
            [
                p.single.calc(federal_agi),
                p.joint.calc(federal_agi),
                p.head_of_household.calc(federal_agi),
                p.separate.calc(federal_agi),
                p.surviving_spouse.calc(federal_agi),
            ],
        )
        return min_(or_federal_income_tax, cap)



from policyengine_us.model_api import *


class ca_renter_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "California Renter Tax Credit"
    unit = USD
    documentation = "https://casetext.com/statute/california-codes/california-revenue-and-taxation-code/division-2-other-taxes/part-10-personal-income-tax/chapter-2-imposition-of-tax/section-170535-credit-for-qualified-renter"
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        # Check eligibility based on state, rent, filing status, and income.
        p = parameters(period).gov.states.ca.tax.income.credits.renter
        agi = tax_unit("ca_agi", period)
        has_rent = add(tax_unit, period, ["rent"]) > 0
        filing_status = tax_unit("filing_status", period)
        income_cap = p.income_cap[filing_status]
        income_eligible = agi <= income_cap
        eligible = income_eligible & has_rent
        # Determine amount if eligible based on filing status.
        amount_if_eligible = p.amount[filing_status]
        # Return eligibility * (amount if eligible).
        return eligible * amount_if_eligible



from policyengine_core.simulations import Simulation


def get_stored_variables(simulation: Simulation) -> list:
    stored_variables = []
    for variable in simulation.tax_benefit_system.variables:
        holder = simulation.get_holder(variable)
        if len(holder.get_known_periods()) > 0:
            stored_variables.append(variable)
    return stored_variables



from policyengine_us.model_api import *


class small_area_fair_market_rent(Variable):
    value_type = float
    entity = Household
    label = "Small area fair market rent"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class non_sch_d_capital_gains(Variable):
    value_type = float
    entity = Person
    label = "Capital gains not reported on Schedule D"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class was_in_foster_care(Variable):
    value_type = bool
    entity = Person
    label = "Person was in the a qualifying foster care institution"
    definition_period = YEAR



from policyengine_us.model_api import *


class is_separated(Variable):
    value_type = bool
    entity = Person
    label = "Separated"
    documentation = "Whether the person is separated from a partner."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/7703"



from policyengine_us.model_api import *


class hi_taxable_income_for_alternative_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "Hawaii eligible capital gains for the alternative tax capital gains"
    )
    unit = USD
    definition_period = YEAR
    defined_for = "hi_alternative_tax_on_capital_gains_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income
        filing_status = tax_unit("filing_status", period)
        # line 1 of Hawaii Tax on Capital Gains Worksheet
        # on Hawaii Resident Income Tax Instructions N-11 Rev.2022, page=33
        taxable_income = tax_unit("hi_taxable_income", period)
        # line 4
        net_lt_capital_gain = add(
            tax_unit, period, ["long_term_capital_gains"]
        )
        # line 7
        net_capital_gain = tax_unit("net_capital_gain", period)
        # line 8
        smaller_net_capital_gain = min_(net_capital_gain, net_lt_capital_gain)
        # The filer is eligible to compute the tax on the smaller of the taxable income less capital gains
        # or the taxable income capped at a threshold below the tax on capital gains rate
        reduced_taxable_income = taxable_income - smaller_net_capital_gain
        cap_single = p.rates.single.thresholds[6]
        cap_joint = p.rates.joint.thresholds[6]
        cap_surviving_spouse = p.rates.surviving_spouse.thresholds[6]
        cap_separate = p.rates.separate.thresholds[6]
        cap_hoh = p.rates.head_of_household.thresholds[6]
        statuses = filing_status.possible_values
        cap = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.JOINT,
                filing_status == statuses.SURVIVING_SPOUSE,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
            ],
            [
                cap_single,
                cap_joint,
                cap_surviving_spouse,
                cap_separate,
                cap_hoh,
            ],
        )
        capped_taxable_income = min_(taxable_income, cap)
        return max_(reduced_taxable_income, capped_taxable_income)



from policyengine_us.model_api import *


class co_low_income_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado Low-income Child Care Expenses Credit"
    unit = USD
    documentation = (
        "https://casetext.com/statute/colorado-revised-statutes/title-39-taxation/specific-taxes/income-tax/article-22-income-tax/part-1-general/section-39-22-1195-child-care-expenses-tax-credit-legislative-declaration-definitions"
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=46"
    )
    definition_period = YEAR
    defined_for = "co_low_income_cdcc_eligible"

    def formula(tax_unit, period, parameters):
        # follow 2022 DR 0347 form and its instructions (in Book cited above):
        p = parameters(period).gov.states.co.tax.income.credits
        # estimate care expenses for just children
        care_expenses = tax_unit("tax_unit_childcare_expenses", period)
        age = tax_unit.members("age", period)
        eligible_kid = age < p.cdcc.low_income.child_age_threshold
        eligible_kids = tax_unit.sum(eligible_kid)
        total_eligibles = tax_unit("count_cdcc_eligible", period)
        eligible_kid_ratio = np.zeros_like(total_eligibles)
        mask = total_eligibles > 0
        eligible_kid_ratio[mask] = eligible_kids[mask] / total_eligibles[mask]
        kid_expenses = care_expenses * eligible_kid_ratio
        capped_kid_expenses = min_(  # Line 3
            kid_expenses, tax_unit("min_head_spouse_earned", period)
        )
        # calculate capped credit amount
        credit = p.cdcc.low_income.rate * capped_kid_expenses  # Line 11
        cap = p.cdcc.low_income.max_amount.calc(eligible_kids)  # Table A
        return min_(credit, cap)  # Line 12



from policyengine_us.model_api import *


class household_income_ami_ratio(Variable):
    value_type = float
    entity = Household
    label = "Ratio of household income to area median income"
    definition_period = YEAR



from policyengine_us.model_api import *


class va_national_guard_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia national guard pay subtraction"
    defined_for = StateCode.VA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.va.tax.income.subtractions.national_guard_pay
        national_guard_pay = add(tax_unit, period, ["military_service_income"])
        return min_(national_guard_pay, p.cap)



from policyengine_us.model_api import *


class de_capped_real_estate_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware capped real estate tax"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenuefiles.delaware.gov/2022/TY22_PIT-RSA_2022-02_PaperInteractive.pdf",  # § 1109
        "https://delcode.delaware.gov/title30/c011/sc02/index.html",
    )
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        filing_status = tax_unit("filing_status", period)
        real_estate_tax = add(tax_unit, period, ["real_estate_taxes"])
        cap = p.itemized.salt_and_real_estate.cap[filing_status]
        return min_(real_estate_tax, cap)



from policyengine_us.model_api import *


class nj_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "New Jersey property tax credit eligibility"
    definition_period = YEAR
    reference = (
        "https://www.nj.gov/treasury/taxation/pdf/other_forms/tgi-ee/2021/1040i.pdf#page=26"
        "https://www.nj.gov/treasury/taxation/pdf/other_forms/tgi-ee/2022/1040i.pdf#page=26"
        "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=27"
    )
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nj.tax.income.credits.property_tax

        # determine if pays property taxes via home ownership or renting
        owner_pays_ptax = add(tax_unit, period, ["real_estate_taxes"]) > 0
        renter_pays_ptax = tax_unit("rents", period)
        pays_ptax = owner_pays_ptax | renter_pays_ptax

        # if filing jointly, only one spouse needs to be 65+ or blind/disabled
        blind_head = tax_unit("blind_head", period)
        disabled_head = tax_unit("disabled_head", period)
        blind_spouse = tax_unit("blind_spouse", period)
        disabled_spouse = tax_unit("disabled_spouse", period)
        senior_head_or_spouse = (
            tax_unit("greater_age_head_spouse", period) >= p.age_threshold
        )
        senior_blind_disabled = (
            blind_head
            | disabled_head
            | blind_spouse
            | disabled_spouse
            | senior_head_or_spouse
        )

        # return eligiblity for property tax credit
        deduction_eligibility = tax_unit(
            "nj_property_tax_deduction_eligible", period
        )
        return deduction_eligibility | (pays_ptax & senior_blind_disabled)



from policyengine_us.model_api import *


class excess_withheld_payroll_tax(Variable):
    value_type = float
    entity = Person
    label = "excess withheld payroll tax"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class misc_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Miscellaneous deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/67#b"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.itemized.misc
        if p.applies:
            expenses = tax_unit("total_misc_deductions", period)
            misc_floor = p.floor * tax_unit("positive_agi", period)
            return max_(0, expenses - misc_floor)
        return 0



from policyengine_us.model_api import *


class in_bonus_depreciation_add_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana bonus depreciation add back"
    definition_period = YEAR
    documentation = "Income (or loss) included in Federal AGI under Section 168(k)'s bonus depreciation less the amount that would have been included without it."
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"  # (a)(15)
    # use federal variables if they are added later
    defined_for = StateCode.IN



from policyengine_us.model_api import *


class la_general_relief_personal_property_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Los Angeles County General Relief based on the personal property value requirements"
    definition_period = YEAR
    # Person has to be a resident of LA County
    defined_for = "in_la"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"

    def formula(spm_unit, period, parameters):
        personal_property = add(spm_unit, period, ["personal_property"])
        p = parameters(period).gov.local.ca.la.general_relief.eligibility.limit
        return personal_property < p.personal_property



from policyengine_us.model_api import *


class taxsim_v10(Variable):
    value_type = float
    entity = TaxUnit
    label = "Federal AGI"
    unit = USD
    documentation = "TAXSIM federal AGI"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit("adjusted_gross_income", period)



from policyengine_us.model_api import *


class ar_inflation_relief_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas inflation relief income-tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR

    adds = ["ar_inflation_relief_credit_person"]



from policyengine_us.model_api import *


class me_step_4_share_of_child_care_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine step 4 share of child care expenses"
    documentation = "Share of child care expenses that are from programs classified as Step 4 in Maine"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://www.mainelegislature.org/legis/statutes/36/title36sec5218.html",
        "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_a_ff.pdf#page=2",
    ]
    defined_for = StateCode.ME



from policyengine_us.model_api import *


class other_net_gain(Variable):
    value_type = float
    entity = TaxUnit
    label = "Other net gains"
    unit = USD
    documentation = "Other net gain/loss from Form 4797"
    definition_period = YEAR



from policyengine_us.model_api import *


class retirement_benefits_from_ss_exempt_employment(Variable):
    value_type = float
    entity = Person
    label = "Retirement benefits amount from SS exempt employment"
    unit = USD
    documentation = "Amount of a recipient receive retirement benefits from SS exempt employment"
    reference = "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=18"
    definition_period = YEAR



from policyengine_us.model_api import *


class ma_maximum_state_supplement(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts maximum State Supplement payment amount"
    unit = USD
    definition_period = YEAR
    defined_for = "is_ssi_eligible"

    def formula(person, period, parameters):
        """
        Returns the maximum annual state supplement for an SSI-eligible individual,
        before subtracting any leftover income. Depends on:
          - The person's state of residence
          - Living arrangement (e.g. FULL_COST, REST_HOME, etc.)
          - Category (Aged, Blind, or Disabled)
          - Single vs. Joint claim (couple)
        """

        # 1. Get state and living arrangement from the person's household.
        living_arrangement = person.household(
            "ma_state_living_arrangement", period
        )

        # 2. Load the state's parameter dictionary. This is structured like:
        # p[living_arrangement][category][("1" or "2")] = monthly rate
        p = parameters(period).gov.states.ma.dta.ssp.amount
        # For example: p["FULL_COST"]["AGED"]["1"] = 128.82

        # 3. Retrieve the sub-dict for this person's state & living arrangement
        # e.g. amounts = p["FULL_COST"]
        amounts_for_state_la = p[living_arrangement]

        # 4. Identify whether the person is aged, blind, or disabled.
        #    (It's possible to be both aged+blind, so we might pick whichever rate is higher.)
        is_blind = person("is_blind", period)
        is_aged = person("is_ssi_aged", period)
        is_disabled = person("is_ssi_disabled", period)

        # 5. We'll build an array of the single or double (couple) label: e.g. "1" or "2".
        #    This matches the structure in the parameter file.
        joint_claim = person("ssi_claim_is_joint", period)
        # Convert True -> "2", False -> "1"
        single_or_double_str = where(joint_claim, 2, 1).astype(str)

        # 6. For vectorization, we note how many people are in the array dimension.
        #    We do the same trick for categories. For each person, we might pick
        #    'AGED', 'BLIND', or 'DISABLED' monthly rate. This code uses np arrays:
        ssi_categories = person("ssi_category", period).possible_values
        count_persons = len(is_blind)  # number of entries in the vector

        # 7. Create arrays for each category's monthly rate. For each person, we do:
        #    amounts_for_state_la["AGED"]["1"], amounts_for_state_la["AGED"]["2"], etc.
        #    Then multiply by the boolean (is_aged, is_blind, is_disabled).
        #    We use max_() to pick whichever is relevant if more than one category is true.

        monthly_aged_rate = (
            amounts_for_state_la[
                np.array([ssi_categories.AGED] * count_persons)
            ][single_or_double_str]
            * is_aged
        )

        monthly_blind_rate = (
            amounts_for_state_la[
                np.array([ssi_categories.BLIND] * count_persons)
            ][single_or_double_str]
            * is_blind
        )

        monthly_disabled_rate = (
            amounts_for_state_la[
                np.array([ssi_categories.DISABLED] * count_persons)
            ][single_or_double_str]
            * is_disabled
        )

        # 8. We'll pick the maximum, so if someone's "is_aged" is True and "is_blind" is False,
        #    the first product is nonzero, second is zero, etc.
        #    If they're both true, we pick whichever is higher.
        monthly_rate = np.maximum.reduce(
            [monthly_aged_rate, monthly_blind_rate, monthly_disabled_rate]
        )

        # 9. Multiply by 12 to get an annual figure. This is the maximum annual supplement
        #    (still for each person individually).
        annual_rate = monthly_rate * MONTHS_IN_YEAR

        # 10. Summation across the marital unit. If we have a couple, we might want
        #     the total combined amount, then split it. Or we might directly keep it as per-person.
        # Here, they sum for the entire marital unit, then later they do dividing if needed.
        combined_amount_for_marital_unit = person.marital_unit.sum(annual_rate)

        # 11. Finally, the code divides by 2 if it's a joint claim, presumably so that
        #     each person sees their half. If it's not joint, it just returns the full amount.
        # This "per-person" approach might be confusing if the state has a special "couple rate"
        # that isn't exactly double or half. But that's how the code is set up now.
        divisor = where(joint_claim, 2, 1)
        return combined_amount_for_marital_unit / divisor



from policyengine_us.model_api import *


class dc_tanf_is_working(Variable):
    value_type = bool
    entity = Person
    label = "Person is working under the work requirement for DC Temporary Assistance for Needy Families (TANF)"
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.19b"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.dc.dhs.tanf.work_requirement.required_hours
        weekly_hours_worked = person("weekly_hours_worked", period.this_year)
        age = person("monthly_age", period)
        # For single parent with a child under 6, work 20 hours pr week
        # For single parent with a child at 6 or older, work 30 hours pr week
        spm_unit = person.spm_unit
        is_youngest_member = person.get_rank(spm_unit, age, spm_unit) == 0
        youngest_age = spm_unit.sum(is_youngest_member * age)
        single_parent_requirement = (
            weekly_hours_worked >= p.single_parent.amount.calc(youngest_age)
        )
        # For 2-parent, work 35 hours pr week (combined)
        # Or 55 hours per week if family receives federally-funded childcare and
        # no adult has a disability (not modeled)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        parent_hours = where(is_head_or_spouse, weekly_hours_worked, 0)
        combined_parent_hours = spm_unit.sum(parent_hours)
        two_parent_requirement = combined_parent_hours >= p.two_parents.amount
        is_two_parent_unit = spm_unit.sum(is_head_or_spouse) > 1
        return where(
            is_two_parent_unit,
            two_parent_requirement,
            single_parent_requirement,
        )



from policyengine_us.model_api import *


class ca_la_infant_supplement(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Los Angeles County infant supplement"
    defined_for = "in_la"

    def formula(household, period, parameters):
        p = parameters(period).gov.local.ca.la.dss.infant_supplement
        person = household.members
        is_in_group_home = household.any(
            person("is_in_foster_care_group_home", period)
        )
        base_amount = where(
            is_in_group_home, p.amount.group_home, p.amount.base
        )
        children = add(household, period, ["is_child"])
        eligible_person = person(
            "ca_la_infant_supplement_eligible_person", period
        )
        eligible_person_present = household.any(eligible_person)
        return base_amount * children * eligible_person_present



from policyengine_us.model_api import *


class snap_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = "SNAP earned income"
    documentation = (
        "Earned income for calculating the SNAP earned income deduction"
    )
    reference = "https://www.law.cornell.edu/cfr/text/7/273.9#b_1"
    unit = USD

    adds = ["snap_earned_income_person"]



from policyengine_us.model_api import *


class az_long_term_capital_gains_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona long-term capital gains subtraction"
    unit = USD
    documentation = "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2022_140i.pdf#page=31"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.subtractions.capital_gains

        long_term_capital_gains = add(
            tax_unit, period, ["long_term_capital_gains"]
        )

        return max_(0, long_term_capital_gains) * p.rate



from policyengine_us.model_api import *


class pell_grant_head_available_income(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Pell Grant head available income"
    definition_period = YEAR

    def formula(person, period, parameters):
        allowances = person("pell_grant_head_allowances", period)
        income = person.tax_unit("pell_grant_primary_income", period)
        formula = person("pell_grant_formula", period)
        p = parameters(period).gov.ed.pell_grant.head
        return (income - allowances) * p.income_assessment_rate[formula]



from policyengine_us.model_api import *


class ny_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY exemptions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/616"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        count_dependents = tax_unit("tax_unit_child_dependents", period)
        dependent_exemption = parameters(
            period
        ).gov.states.ny.tax.income.exemptions.dependent
        return dependent_exemption * count_dependents



from policyengine_us.model_api import *


class ma_tafdc_eligible_infant(Variable):
    value_type = bool
    entity = Person
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) eligible infant"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-705-600"
    )
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        # Monthly age is used to determine the age of the infant
        age = person("monthly_age", period)
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.tafdc.eligibility.age_limit
        return age < p.infant



from policyengine_us.model_api import *


class md_unmarried_childless_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland unmarried childless refundable EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=19"
    defined_for = "md_qualifies_for_unmarried_childless_eitc"

    def formula(tax_unit, period, parameters):
        md_tax = tax_unit("md_income_tax_before_credits", period)
        md_unmarried_childless_non_refundable_eitc = tax_unit(
            "md_unmarried_childless_non_refundable_eitc", period
        )
        return max_(md_unmarried_childless_non_refundable_eitc - md_tax, 0)



from policyengine_us.model_api import *


class free_school_meals(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "free school meals"
    unit = USD
    documentation = "Value of free school meals."

    def formula(spm_unit, period, parameters):
        disabled_programs = parameters(period).simulation.disabled_programs
        if "free_school_meals" in disabled_programs:
            return spm_unit("free_school_meals_reported", period)
        tier = spm_unit("school_meal_tier", period)
        is_free = tier == tier.possible_values.FREE
        return is_free * spm_unit("school_meal_net_subsidy", period)



from policyengine_us.model_api import *


class ky_agi(Variable):
    value_type = float
    entity = Person
    label = "Kentucky adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://revenue.ky.gov/Forms/740%20Packet%20Instructions%205-9-23.pdf#page=11"
    defined_for = StateCode.KY

    adds = [
        "ky_additions",
        "adjusted_gross_income_person",
    ]
    subtracts = ["ky_subtractions"]



from policyengine_us.model_api import *


class az_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona standard deduction"
    unit = USD
    documentation = "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01041.htm"
    definition_period = YEAR
    defined_for = StateCode.AZ

    adds = [
        "az_base_standard_deduction",
        "az_increased_standard_deduction_for_charitable_contributions",
    ]



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_aca_ptc_additional_bracket() -> Reform:
    class aca_ptc_phase_out_rate(Variable):
        """
        ACA Premium Tax Credit phase-out rate with linear bracket extension.

        This reform extends premium subsidies beyond the standard 400% FPL cliff
        using a linear formula. Above the transition threshold (typically 300% FPL),
        the contribution percentage increases linearly based on the increment rate
        (typically 4 percentage points per 100% FPL).

        The reform creates a more gradual phase-out of subsidies, reducing the
        cliff effect where households just above 400% FPL lose all subsidies.

        Related issue: https://github.com/PolicyEngine/policyengine-us/issues/6629
        """

        value_type = float
        entity = TaxUnit
        label = "ACA PTC phase-out rate with additional bracket (i.e., IRS Form 8962 'applicable figure')"
        unit = "/1"
        definition_period = YEAR
        reference = [
            "26 U.S. Code § 36B(b)(3)(A) - Refundable credit for coverage under a qualified health plan",
            "https://www.law.cornell.edu/uscode/text/26/36B#b_3_A",
        ]

        def formula(tax_unit, period, parameters):
            magi_frac = tax_unit("aca_magi_fraction", period)
            p = parameters(
                period
            ).gov.contrib.aca.ptc_additional_bracket.brackets
            return np.interp(magi_frac, p.thresholds, p.amounts)

    class is_aca_ptc_eligible(Variable):
        """
        Extended ACA PTC eligibility beyond 400% FPL.
        """

        value_type = bool
        entity = Person
        label = "Person is eligible for ACA premium tax credit and pays ACA premium"
        definition_period = YEAR

        def formula(person, period, parameters):
            # determine status eligibility for ACA PTC
            fstatus = person.tax_unit("filing_status", period)
            separate = fstatus == fstatus.possible_values.SEPARATE
            immigration_eligible = person(
                "is_aca_ptc_immigration_status_eligible", period
            )
            taxpayer_has_itin = person.tax_unit("taxpayer_has_itin", period)
            is_status_eligible = (
                taxpayer_has_itin & ~separate & immigration_eligible
            )

            # determine coverage eligibility for ACA plan
            INELIGIBLE_COVERAGE = [
                "is_medicaid_eligible",
                "is_chip_eligible",
                "is_aca_eshi_eligible",
                "is_medicare_eligible",
            ]
            is_coverage_eligible = (
                add(person, period, INELIGIBLE_COVERAGE) == 0
            )

            # determine income eligibility for ACA PTC (using reform parameter)
            p = parameters(period).gov.contrib.aca.ptc_additional_bracket
            magi_frac = person.tax_unit("aca_magi_fraction", period)
            is_income_eligible = p.income_eligibility.calc(magi_frac)

            # determine which people pay an age-based ACA plan premium
            p_aca = parameters(period).gov.aca
            is_aca_adult = person("age", period) > p_aca.slcsp.max_child_age
            child_pays = (
                person("aca_child_index", period) <= p_aca.max_child_count
            )
            pays_aca_premium = is_aca_adult | child_pays

            return (
                is_status_eligible
                & is_coverage_eligible
                & is_income_eligible
                & pays_aca_premium
            )

    class reform(Reform):
        def apply(self):
            self.update_variable(aca_ptc_phase_out_rate)
            self.update_variable(is_aca_ptc_eligible)

    return reform


def create_aca_ptc_additional_bracket_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_aca_ptc_additional_bracket()

    p = parameters.gov.contrib.aca.ptc_additional_bracket
    current_period = period_(period)

    # Check if reform is active within a 5-year lookahead window
    # This allows the reform to be selected in the web app interface
    # even if it's scheduled to start in a future year
    reform_active = False
    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_aca_ptc_additional_bracket()
    else:
        return None


aca_ptc_additional_bracket = create_aca_ptc_additional_bracket_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class co_ccap_smi_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Colorado Child Care Assistance Program state median income-based income eligibility test"
    reference = "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=19"
    definition_period = MONTH
    # defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        monthly_gross_income = np.round(
            spm_unit("co_ccap_countable_income", period),
            2,
        )
        year = period.start.year
        if period.start.month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p = parameters(instant_str).gov.states.co.ccap
        # Calculate monthly smi limit
        smi = spm_unit("co_ccap_smi", period)
        income_limit = np.round(smi * p.entry.smi_rate, 2)
        return monthly_gross_income < income_limit



from policyengine_us.model_api import *


class md_capital_gains_surtax_applies(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Maryland capital gains surtax applies"
    definition_period = YEAR
    reference = [
        "https://mgaleg.maryland.gov/2025RS/Chapters_noln/CH_604_hb0352e.pdf#page=162"  # Maryland House Bill 352 - Budget Reconciliation and Financing Act of 2025
    ]
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.capital_gains

        # First check if surtax is enabled for this year
        if not p.surtax_applies:
            return False

        # Check Maryland adjusted gross income threshold
        md_agi = tax_unit("md_agi", period)

        return md_agi > p.surtax_threshold



from policyengine_us.model_api import *


class heating_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit heating cost"
    unit = USD
    definition_period = YEAR

    adds = ["heating_expense_person"]



from policyengine_us.model_api import *


class oh_tax_before_joint_filing_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio tax liability before the joint filing credit"
    reference = (
        # 2021 Ohio Schedule of Credits
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/sch-cre.pdf",
        # 2022 Ohio Schedule of Credits
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/itschedule-credits.pdf",
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        tax_before_credits = tax_unit(
            "oh_income_tax_before_non_refundable_credits", period
        )
        total_credits = tax_unit("oh_partial_non_refundable_credits", period)
        return max_(0, tax_before_credits - total_credits)



from policyengine_us.model_api import *


class TXCCSProviderType(Enum):
    LCCC = "Licensed Child Care Center"
    LCCH = "Licensed Child Care Home"
    RCCH = "Registered Child Care Home"
    RELATIVE = "Relative Care"


class tx_ccs_provider_type(Variable):
    value_type = Enum
    possible_values = TXCCSProviderType
    default_value = TXCCSProviderType.LCCC
    entity = Person
    definition_period = MONTH
    label = "Texas Child Care Services (CCS) provider type"
    defined_for = StateCode.TX
    reference = "https://www.twc.texas.gov/sites/default/files/ccel/docs/bcy25-board-max-provider-payment-rates-4-age-groups-twc.pdf"



from policyengine_us.model_api import *


class il_cta_reduced_fare_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Illinois Chicago Transit Authority Reduced Fare Program"
    definition_period = YEAR
    defined_for = StateCode.IL
    reference = "https://www.transitchicago.com/reduced-fare-programs/"

    def formula(person, period, parameters):
        children_reduced_fare = person(
            "il_cta_children_reduced_fare_eligible", period
        )
        rta_reduced_fare = person("il_cta_rta_reduced_fare_eligible", period)
        student_reduced_fare = person(
            "il_cta_student_reduced_fare_eligible", period
        )
        return children_reduced_fare | rta_reduced_fare | student_reduced_fare



from policyengine_us.model_api import *


class nc_tanf_countable_gross_unearned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina TANF countable gross unearned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.nc.ncdhhs.tanf.income
        # Sum unearned sources, plus child support if not currently enrolled.
        gross_unearned = add(spm_unit, period, p.unearned)
        child_support = add(spm_unit, period, ["child_support_received"])
        tanf_enrolled = spm_unit("is_tanf_enrolled", period)
        additional_amount = where(tanf_enrolled, 0, child_support)
        return gross_unearned + additional_amount



from policyengine_us.model_api import *


class months_receiving_social_security_disability(Variable):
    value_type = int
    entity = Person
    label = "Number of months person has received social security disability"
    definition_period = YEAR



from policyengine_us.model_api import *


class is_medicaid_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for Medicaid"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/42/1396a#a_10"
        "https://www.kff.org/racial-equity-and-health-policy/fact-sheet/key-facts-on-health-coverage-of-immigrants"
    )

    def formula(person, period, parameters):
        category = person("medicaid_category", period)
        categorically_eligible = category != category.possible_values.NONE
        immigration_status_eligible = person(
            "is_medicaid_immigration_status_eligible", period
        )
        ca_ffyp_eligible = person("ca_ffyp_eligible", period)
        p = parameters(period).gov.hhs.medicaid.eligibility
        if p.work_requirements.applies:
            work_requirement_eligible = person(
                "medicaid_work_requirement_eligible", period
            )
            return (
                categorically_eligible
                & immigration_status_eligible
                & work_requirement_eligible
            ) | ca_ffyp_eligible
        return (
            categorically_eligible & immigration_status_eligible
        ) | ca_ffyp_eligible



from policyengine_us.model_api import *


class ky_family_size_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kentucky family size tax credit"
    unit = "/1"
    definition_period = YEAR
    reference = (
        "https://apps.legislature.ky.gov/law/statutes/statute.aspx?id=49188"
    )
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        rate = tax_unit("ky_family_size_tax_credit_rate", period)
        income = tax_unit(
            "ky_income_tax_before_non_refundable_credits_unit", period
        )
        personal_credits = tax_unit("ky_personal_tax_credits", period)
        reduced_income = max_(income - personal_credits, 0)

        return rate * reduced_income



from policyengine_us.model_api import *


class ca_capi_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "California CAPI eligible"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "https://www.cdss.ca.gov/Portals/9/CAPI/CAPI_Regulations-Accessible.pdf"

    def formula(spm_unit, period, parameters):
        resource_eligible = spm_unit("ca_capi_resource_eligible", period)
        income_eligible = spm_unit("ca_capi_income_eligible", period)
        has_eligible_person = (
            add(spm_unit, period, ["ca_capi_eligible_person"]) > 0
        )
        return resource_eligible & income_eligible & has_eligible_person



from policyengine_us.model_api import *


class OvertimeExemptionCategory(Enum):
    NONE = "No exemption category"
    NEVER_WORKED = "Never worked"
    MILITARY = "Military"
    EXECUTIVE_ADMINISTRATIVE = "Executive, administrative, or professional"
    FARMER_FISHER = "Farmer or fisher"
    COMPUTER_SCIENTIST = "Computer scientist"


class fsla_overtime_occupation_exemption_category(Variable):
    value_type = Enum
    possible_values = OvertimeExemptionCategory
    default_value = OvertimeExemptionCategory.NONE
    entity = Person
    label = "FSLA occupation categories for overtime exemption"
    reference = "https://www.law.cornell.edu/uscode/text/29/213"
    definition_period = YEAR

    def formula(person, period, parameters):
        # Using select with conditions and values for vectorized operation
        return select(
            [
                person("is_military", period),
                person("has_never_worked", period),
                person("is_executive_administrative_professional", period),
                person("is_farmer_fisher", period),
                person("is_computer_scientist", period),
                True,  # Default case
            ],
            [
                OvertimeExemptionCategory.MILITARY,
                OvertimeExemptionCategory.NEVER_WORKED,
                OvertimeExemptionCategory.EXECUTIVE_ADMINISTRATIVE,
                OvertimeExemptionCategory.FARMER_FISHER,
                OvertimeExemptionCategory.COMPUTER_SCIENTIST,
                OvertimeExemptionCategory.NONE,
            ],
        )



from policyengine_us.model_api import *


class is_child_dependent(Variable):
    value_type = bool
    entity = Person
    label = "Is a child dependent based on the IRS definition"
    reference = "https://www.law.cornell.edu/uscode/text/26/152#c_3_A_ii"
    definition_period = YEAR
    defined_for = "is_tax_unit_dependent"

    def formula(person, period, parameters):
        age = person("age", period)
        p = parameters(period).gov.irs.dependent.ineligible_age
        student = person("is_full_time_student", period)
        age_limit = where(student, p.student, p.non_student)
        return age < age_limit



from policyengine_us.model_api import *


class hi_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.HI

    adds = "gov.states.hi.tax.income.credits.refundable"



from policyengine_us.model_api import *


class la_retirement_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "Louisiana retirement exemption for each person"
    unit = USD
    definition_period = YEAR
    reference = "https://www.legis.la.gov/legis/Law.aspx?d=102133"
    defined_for = StateCode.LA

    def formula(person, period, parameters):
        pension_income = person("taxable_pension_income", period)
        age = person("age", period)
        p = parameters(
            period
        ).gov.states.la.tax.income.exempt_income.retirement
        cap = p.cap.calc(age)
        deductible_pensions = min_(pension_income, cap)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return deductible_pensions * is_head_or_spouse



from policyengine_us.model_api import *


class nm_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/856ebf4b-3814-49dd-8631-ebe579d6a42b/Personal%20Income%20Tax.pdf"  # 7-2-18.15
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        eligible = tax_unit("nm_eitc_eligible", period)
        maximum = tax_unit("eitc_maximum", period)
        phased_in = tax_unit("eitc_phased_in", period)
        reduction = tax_unit("eitc_reduction", period)
        limitation = max_(0, maximum - reduction)
        eitc = eligible * min_(phased_in, limitation)
        rate = parameters(period).gov.states.nm.tax.income.credits.eitc.match
        return eitc * rate



from policyengine_us.model_api import *


class tx_ccs_work_exempt(Variable):
    value_type = bool
    entity = Person
    label = "Texas CCS work requirement exempt"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/regulations/texas/40-Tex-Admin-Code-SS-809-50"
    defined_for = StateCode.TX

    def formula(person, period):
        # Exemption 1: Full-time student
        is_full_time_student = person("is_full_time_student", period)

        # Exemption 2: Parent with documented medical disability
        is_disabled = person("is_disabled", period)

        # Exemption 3: Parent caring for a physically or mentally disabled family member
        # This exemption is NOT implemented due to complexity in determining:
        # - Which parent is the designated caregiver
        # - Whether caring for disabled spouse vs disabled child has different implications
        # - Whether both parents can claim exemption or just one
        # Future enhancement could add this with better data on caregiving roles

        return is_full_time_student | is_disabled



from policyengine_us.model_api import *


class ma_eaedc_eligible_disabled_head_or_spouse(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Disabled head or spouse present for Massachusetts EAEDC"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-010"  # (B)

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        is_disabled = person("is_disabled", period)
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        ssi_eligible = person("is_ssi_eligible", period)
        return spm_unit.any(is_disabled & head_or_spouse & ~ssi_eligible)



from policyengine_us.model_api import *


class ok_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        # follows Schedule 511-D in references:
        # ... calculate pre-limit OK itemized deductions
        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        capped_property_taxes = tax_unit("capped_property_taxes", period)
        ok_itm_deds = itm_deds_less_salt + capped_property_taxes
        # ... apply partial limit on OK itemized deductions
        EXEMPT_ITEMS = [
            "medical_expense_deduction",
            "charitable_deduction",
        ]
        exempt_deds = add(tax_unit, period, EXEMPT_ITEMS)
        net_deds = max_(0, ok_itm_deds - exempt_deds)
        p = parameters(period).gov.states.ok.tax.income
        limited_net_deds = min_(net_deds, p.deductions.itemized.limit)
        return exempt_deds + limited_net_deds



from policyengine_us.model_api import *


class wa_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington refundable tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WA

    adds = "gov.states.wa.tax.income.credits.refundable"



from policyengine_us.model_api import *


class keogh_distributions(Variable):
    value_type = float
    entity = Person
    label = "Keogh plan distributions"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class dc_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC Temporary Assistance for Needy Families (TANF) due to resources"
    definition_period = MONTH
    reference = "https://dhs.dc.gov/sites/default/files/dc/sites/dhs/service_content/attachments/DC%20TANF%20State%20Plan_Oct-2023.pdf#page=40"
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.tanf.resource_limit
        person = spm_unit.members
        # Check if the household has at least one elderly member.
        has_elderly = spm_unit.any(
            person("monthly_age", period) >= p.higher.age_threshold
        )
        # Check if the household has at least one disabled member.
        has_disabled = spm_unit.any(person("is_disabled", period))
        # Look up resource limit by the condition.
        resource_limit = where(
            has_elderly | has_disabled,
            p.higher.amount,
            p.lower.amount,
        )
        countable_resources = spm_unit("dc_tanf_countable_resources", period)
        return countable_resources <= resource_limit



from policyengine_us.model_api import *


class il_aabd_countable_assets(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = (
        "Illinois Aid to the Aged, Blind or Disabled (AABD) countable assets"
    )
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.140",
    )
    defined_for = StateCode.IL

    adds = "gov.states.il.dhs.aabd.asset.sources"



from policyengine_us.model_api import *


class ny_college_tuition_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY college tuition credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (t)
    defined_for = "ny_college_tuition_credit_eligible"

    def formula(tax_unit, period, parameters):
        tuition = tax_unit("ny_allowable_college_tuition_expenses", period)
        p = parameters(period).gov.states.ny.tax.income.credits.college_tuition
        # Apply a tiered rate structure.
        amount_pre_percentage = p.rate.calc(tuition)
        return amount_pre_percentage * p.applicable_percentage



from policyengine_us.model_api import *


class amt_income_less_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax Income less exemptions"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) income less exemptions"

    def formula(tax_unit, period, parameters):
        # Form 6251, Part I
        # Line 4
        amt_income = tax_unit("amt_income", period)
        # For filers subject to the kiddie tax, the deductions are not added back
        taxable_income = tax_unit("taxable_income", period)
        kiddie_tax_applies = tax_unit("amt_kiddie_tax_applies", period)
        applied_income = where(kiddie_tax_applies, taxable_income, amt_income)
        # Form 6251, Part II top
        p = parameters(period).gov.irs.income.amt
        phase_out = p.exemption.phase_out
        filing_status = tax_unit("filing_status", period)
        # Line 5 the exemption amount is based on filing status and
        # is phased-out at higher income
        base_exemption_amount = p.exemption.amount[filing_status]
        income_excess = max_(0, amt_income - phase_out.start[filing_status])
        exemption_phase_out = phase_out.rate * income_excess
        reduced_exemption_amount = max_(
            0,
            (base_exemption_amount - exemption_phase_out),
        )
        # A reduced exemption amount is applied to kiddie tax filers
        adj_earnings = tax_unit("filer_adjusted_earnings", period)
        child_amount = p.exemption.child.amount

        exemption_cap = where(
            kiddie_tax_applies,
            adj_earnings + child_amount,
            np.inf,
        )
        capped_exemption_amount = min_(reduced_exemption_amount, exemption_cap)
        # Line 6
        return max_(0, applied_income - capped_exemption_amount)



from policyengine_us.model_api import *


class has_marketplace_health_coverage(Variable):
    value_type = bool
    entity = Person
    label = "Is eligible for health insurance from an ACA Marketplace plan because has no employer-sponsored health insurance coverage."
    definition_period = YEAR
    default_value = True



from policyengine_us.model_api import *


class tax_unit_earned_income_last_year(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit earned income last year"
    unit = USD
    definition_period = YEAR

    formula = sum_among_non_dependents("earned_income_last_year")



from policyengine_us.model_api import *


class is_computer_scientist(Variable):
    value_type = bool
    entity = Person
    label = "is employed in a computer science related occupation"
    reference = "https://www.law.cornell.edu/uscode/text/29/213"
    definition_period = YEAR



from policyengine_us.model_api import *


class slcsp_family_tier_multiplier(Variable):
    value_type = float
    entity = TaxUnit
    label = "ACA family tier multiplier for premium calculation"
    unit = "/1"
    definition_period = MONTH
    defined_for = "slcsp_family_tier_applies"

    def formula(tax_unit, period, parameters):
        family_category = tax_unit("slcsp_family_tier_category", period)
        p = parameters(period).gov.aca.family_tier_ratings
        state_code = tax_unit.household("state_code", period)
        in_ny = state_code == state_code.possible_values.NY
        return where(in_ny, p.ny[family_category], p.vt[family_category])



from policyengine_us.model_api import *


class is_rural(Variable):
    value_type = bool
    entity = Household
    definition_period = YEAR
    label = "Is in a rural area"
    reference = "https://www.law.cornell.edu/cfr/text/47/54.505#b_3_i"



from policyengine_us.model_api import *


class hi_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii child and dependent care credit"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = "https://files.hawaii.gov/tax/legal/hrs/hrs_235.pdf#page=40"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.credits.cdcc
        # Schedule X PART II:
        # line 22 = min (line 19, line 20) = return of hi_dependent_care_benefits
        # min(line 23, line 24) = return of hi_cdcc_min_head_spouse_earned
        # line 25 = min(line 22, line 23, line 24):
        # The smaller of the minimum earnings or dependent care benefits is used for the final calculation
        min_earned = min_(
            tax_unit("hi_dependent_care_benefits", period),
            tax_unit("hi_cdcc_min_head_spouse_earned", period),
        )
        # line 26:
        agi = tax_unit("hi_agi", period)
        # line 27:
        rate = p.rate.calc(agi, right=True)
        # return: line 28 = line 27 * line 25:
        return rate * min_earned



from policyengine_us.model_api import *


class il_aabd(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) cash benefit"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.IL

    adds = ["il_aabd_person"]



from policyengine_us.model_api import *


class co_ccap_add_on_parent_fee(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado Child Care Assistance Program add on parent fee"
    reference = (
        "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=41",
        "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=62",
    )
    unit = USD
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        year = period.start.year
        if period.start.month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p = parameters(instant_str).gov.states.co.ccap
        # Calculate base parent fee and add on parent fee.
        gross_income = spm_unit("co_ccap_countable_income", period)
        # snap_fpg is monthly.
        snap_fpg = spm_unit("snap_fpg", period)
        eligible_children = spm_unit("co_ccap_eligible_children", period)
        # Calculate add-on parent fee based on the number of eligible
        # children in a household and income:
        # When income <= fpg: 0
        # When income > fpg: 15 for each additional child
        add_on_parent_fee_amount = (
            eligible_children - 1
        ) * p.parent_fee.add_on
        add_on_parent_fee_applies = gross_income > snap_fpg
        return add_on_parent_fee_amount * add_on_parent_fee_applies



from policyengine_us.model_api import *


def create_increase_taxable_earnings_for_social_security() -> Reform:
    class taxable_earnings_for_social_security(Variable):
        value_type = float
        entity = Person
        label = "Taxable gross earnings for OASDI FICA"
        definition_period = YEAR
        unit = USD

        def formula(person, period, parameters):
            earnings = person("payroll_tax_gross_wages", period)
            p = parameters(period).gov
            base_taxable = min_(earnings, p.irs.payroll.social_security.cap)
            secondary_taxable = max_(
                0,
                earnings - p.contrib.cbo.payroll.secondary_earnings_threshold,
            )
            return base_taxable + secondary_taxable

    class reform(Reform):
        def apply(self):
            self.update_variable(taxable_earnings_for_social_security)

    return reform


def create_increase_taxable_earnings_for_social_security_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_increase_taxable_earnings_for_social_security()

    p = parameters(period).gov.contrib.cbo.payroll

    if p.secondary_earnings_threshold < np.inf:
        return create_increase_taxable_earnings_for_social_security()
    else:
        return None


increase_taxable_earnings_for_social_security = (
    create_increase_taxable_earnings_for_social_security_reform(
        None, None, bypass=True
    )
)



from policyengine_us.model_api import *


class ok_ptc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma property tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/538-H-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/538-H.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ok.tax.income.credits.property_tax
        # determine eligibility
        elderly_head = tax_unit("age_head", period) >= p.age_minimum
        elderly_spouse = tax_unit("age_spouse", period) >= p.age_minimum
        disabled_head = tax_unit("head_is_disabled", period)
        unit_eligible = elderly_head | elderly_spouse | disabled_head
        income = tax_unit("ok_gross_income", period)
        income_eligible = income <= p.income_limit
        eligible = unit_eligible & income_eligible
        # calculate credit if eligible
        tax = add(tax_unit, period, ["real_estate_taxes"])
        excess_property_tax = max_(0, tax - p.income_fraction * income)
        return eligible * min_(p.maximum_credit, excess_property_tax)



from policyengine_us.model_api import *


class sc_tuition_credit(Variable):
    value_type = float
    entity = Person
    label = "South Carolina Tuition Credit"
    defined_for = "sc_tuition_credit_eligible"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/I319_2021.pdf#page=2",
        "https://www.scstatehouse.gov/code/t12c006.php",
        # South Carolina Legal Code | SECTION 12-6-3385 (A)
    )



from policyengine_us.model_api import *


class va_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia Earned Income Tax Credit"
    unit = USD
    documentation = "Refundable or non-refundable Virginia EITC"
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=32"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        refundable_eitc = tax_unit("va_refundable_eitc", period)
        non_refundable_eitc = tax_unit("va_non_refundable_eitc", period)
        claims_refundable = tax_unit("va_claims_refundable_eitc", period)
        amount = where(claims_refundable, refundable_eitc, non_refundable_eitc)
        filing_status = tax_unit("filing_status", period)
        # In the case of separated individuals, the EITC amount is prorated
        is_separate = filing_status == filing_status.possible_values.SEPARATE
        person = tax_unit.members
        eitc_person = person("va_eitc_person", period)
        is_head = person("is_tax_unit_head", period)
        head_eitc = tax_unit.sum(eitc_person * is_head)
        return where(is_separate, head_eitc, amount)



from policyengine_us.model_api import *


class az_family_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Arizona Family Tax Credit"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.credits.family_tax_credits
        income = tax_unit("az_agi", period)
        filing_status = tax_unit("az_filing_status", period)
        status = filing_status.possible_values
        dependents = tax_unit("tax_unit_dependents", period)
        income_limit = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.SEPARATE,
            ],
            [
                p.income_limit.single,
                p.income_limit.joint.calc(dependents),
                p.income_limit.head_of_household.calc(dependents),
                p.income_limit.separate,
            ],
        )
        return income <= income_limit



from policyengine_us.model_api import *


class il_tanf_payment_level_for_initial_eligibility(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Temporary Assistance for Needy Families (TANF) payment level for initial eligibility"
    unit = USD
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.251"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dhs.tanf.payment_level
        fpg = spm_unit("il_tanf_assistance_unit_fpg", period)

        return p.rate * fpg



from policyengine_us.model_api import *


class md_senior_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Maryland Senior Tax Credit"
    definition_period = YEAR
    reference = "https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf#page=15"
    defined_for = StateCode.MD

    def formula_2022(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.credits.senior_tax

        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)

        return agi < p.income_threshold[filing_status]



from policyengine_us.model_api import *


class mo_ptc_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri property tax credit gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-PTS_2021.pdf",
        "https://dor.mo.gov/forms/4711_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.010&bid=6435&hl=property+tax+credit%u2044",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        # line numbers in comments below refer to 2021 Form MO-PTS
        person = tax_unit.members
        # compute core person-level income included in PTC gross income
        sources = [
            "mo_adjusted_gross_income",  # line 1
            "tax_exempt_pension_income",  # line 3
            "tax_exempt_interest_income",  # line 3
        ]
        core_income = tax_unit.sum(add(person, period, sources))
        # compute social security included in PTC gross income (line 2)
        exempt_socsec_benefits = tax_unit("tax_exempt_social_security", period)
        # compute veterans benefits included in PTC gross income (line 5)
        veterans_benefits = person("veterans_benefits", period)
        exclude = person("is_fully_disabled_service_connected_veteran", period)
        included_veterans_benefits = ~exclude * veterans_benefits
        veterans_benefits_income = tax_unit.sum(included_veterans_benefits)
        # compute person-level public assistance included in PTC gross income
        pa_sources = parameters(
            period
        ).gov.states.mo.tax.income.credits.property_tax.public_assistance_types
        public_assistance_income = tax_unit.sum(
            add(person, period, pa_sources)  # line 6
        )
        # compute nonbusiness capital losses in PTC gross income (line 7)
        nonbusiness_losses = tax_unit("limited_capital_loss", period)
        return (  # line 8
            core_income
            + exempt_socsec_benefits
            + veterans_benefits_income
            + public_assistance_income
            + nonbusiness_losses
        )



from policyengine_us.model_api import *


class filer_meets_eitc_identification_requirements(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Filer meets EITC identification requirements"
    reference = (
        # IRC 32(c)(1)(E) specifies filer and spouse needs a taxpayer identification number.
        "https://www.law.cornell.edu/uscode/text/26/32#c_1_E",
    )

    def formula(tax_unit, period, parameters):
        # Both head and spouse in the tax unit must have valid SSN card type to be eligible for the EITC
        person = tax_unit.members
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        eligible_ssn_card_type = person(
            "meets_eitc_identification_requirements", period
        )
        ineligible_head_or_spouse = is_head_or_spouse & ~eligible_ssn_card_type
        return tax_unit.sum(ineligible_head_or_spouse) == 0



from policyengine_us.model_api import *


class ca_itemized_deductions_pre_limitation(Variable):
    value_type = float
    entity = TaxUnit
    label = "California pre-limitation itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ftb.ca.gov/forms/2021/2021-540-ca-instructions.html"
        "https://www.ftb.ca.gov/forms/2022/2022-540-ca-instructions.html"
    )
    defined_for = StateCode.CA

    adds = [
        "ca_investment_interest_deduction",
        "real_estate_taxes",
        "itemized_deductions_less_salt",
    ]



from policyengine_us.model_api import *


class hi_military_pay_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii military reserve or national guard duty pay exclusion"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=13"
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.hi.tax.income.subtractions.military_pay
        person = tax_unit.members
        military_service_income = person("military_service_income", period)
        capped_military_income = min_(military_service_income, p.cap)
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        return tax_unit.sum(capped_military_income * head_or_spouse)



from policyengine_us.model_api import *


class ct_income_tax_recapture(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut income tax recapture"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CT

    adds = [
        "ct_income_tax_low_tax_recapture",
        "ct_income_tax_middle_tax_recapture",
        "ct_income_tax_high_tax_recapture",
    ]



from policyengine_us.model_api import *


class technical_institution_student(Variable):
    value_type = bool
    entity = Person
    label = "Is a technical institution student"
    definition_period = YEAR



from policyengine_us.model_api import *


class school_meal_fpg_ratio(Variable):
    value_type = float
    label = "School meal FPG ratio"
    entity = SPMUnit
    definition_period = YEAR
    documentation = "SPM unit's federal poverty ratio for school meal program"
    unit = "/1"

    def formula(spm_unit, period, parameters):
        income = spm_unit("school_meal_countable_income", period)
        return income / spm_unit("spm_unit_fpg", period)



from policyengine_us.model_api import *


class ca_riv_general_relief_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Riverside County General Relief"
    definition_period = MONTH
    defined_for = "in_riv"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.ca.riv.general_relief
        age = spm_unit.members("monthly_age", period)
        # Adults with dependent children under certain age are NOT eligible for GA.
        # They should apply for CalWORKs (California's TANF) instead.
        age_eligible = spm_unit.all(age >= p.age_threshold)

        immigration_status_eligible = (
            add(
                spm_unit,
                period,
                ["ca_riv_general_relief_immigration_status_eligible"],
            )
            > 0
        )
        meets_work_requirements = (
            add(
                spm_unit,
                period,
                ["ca_riv_general_relief_meets_work_requirements"],
            )
            > 0
        )
        property_eligible = spm_unit(
            "ca_riv_general_relief_property_eligible", period
        )
        income_eligible = spm_unit(
            "ca_riv_general_relief_income_eligible", period
        )

        return (
            age_eligible
            & immigration_status_eligible
            & property_eligible
            & income_eligible
            & meets_work_requirements
        )



from policyengine_us.model_api import *


class home_energy_audit_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Expenditures on home energy audits"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class ut_at_home_parent_credit_earned_income_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Utah at-home parent credit income based on the earned income"
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = (
        "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1005.html",
        "https://www.taxformfinder.org/forms/2021/2021-utah-tc-40-full-packet.pdf#page=23",
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ut.tax.income.credits.at_home_parent

        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        earned_income_eligible = (
            person("earned_income", period) <= p.parent_max_earnings
        )

        return head_or_spouse & earned_income_eligible



from policyengine_us.model_api import *


class ca_mental_health_services_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "CA mental health services tax"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/Search/Home/Confirmation"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("ca_taxable_income", period)
        rate = parameters(
            period
        ).gov.states.ca.tax.income.mental_health_services
        return rate.calc(taxable_income)



from policyengine_us.model_api import *


class is_lifeline_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Lifeline"
    documentation = "Eligible for Lifeline phone or broadband subsidy"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/47/54.409"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.fcc.lifeline
        household = spm_unit.household
        is_on_tribal_land = household("is_on_tribal_land", period)
        non_tribal_lifeline_programs = add(
            spm_unit, period, p.categorical_eligibility
        )
        tribal_lifeline_programs = add(
            spm_unit, period, p.tribal_categorical_eligibility
        )
        categorically_eligible = np.where(
            is_on_tribal_land,
            np.any(tribal_lifeline_programs),
            np.any(non_tribal_lifeline_programs),
        )
        # Use the new unified income eligibility variable
        income_eligible = spm_unit("is_lifeline_income_eligible", period)
        return categorically_eligible | income_eligible



from policyengine_us.model_api import *


class ma_part_a_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part A AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.mass.gov/info-details/mass-general-laws-c62-ss-2"  # (c)
    )
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        part_a_gross_income = tax_unit("ma_part_a_gross_income", period)
        short_term_capital_gains = add(
            tax_unit, period, ["short_term_capital_gains"]
        )
        short_term_capital_loss = max_(0, -short_term_capital_gains)
        nonnegative_short_term_capital_gains = max_(
            0, short_term_capital_gains
        )
        interest_and_dividends = add(
            tax_unit, period, ["taxable_interest_income", "dividend_income"]
        )

        tax = parameters(period).gov.states.ma.tax.income
        interest_dividends_deduction_cap = (
            tax.capital_gains.deductible_against_interest_dividends
        )
        short_term_loss_against_interest_dividends = min_(
            interest_dividends_deduction_cap,
            min_(
                interest_and_dividends,
                short_term_capital_loss,
            ),
        )
        long_term_capital_gains = add(
            tax_unit, period, ["long_term_capital_gains"]
        )
        long_term_capital_loss = max_(0, -long_term_capital_gains)

        long_term_loss_against_short_term_gain = min_(
            long_term_capital_loss,
            nonnegative_short_term_capital_gains,
        )
        remaining_long_term_loss = (
            long_term_capital_loss - long_term_loss_against_short_term_gain
        )
        remaining_interest_dividends = (
            interest_and_dividends - short_term_loss_against_interest_dividends
        )
        remaining_interest_deduction_cap = (
            interest_dividends_deduction_cap
            - short_term_loss_against_interest_dividends
        )
        long_term_loss_against_interest_dividends = min_(
            remaining_interest_deduction_cap,
            min_(
                remaining_long_term_loss,
                remaining_interest_dividends,
            ),
        )

        long_term_capital_gains_on_collectibles = add(
            tax_unit, period, ["long_term_capital_gains_on_collectibles"]
        )
        nonnegative_long_term_capital_gains_on_collectibles = max_(
            0, long_term_capital_gains_on_collectibles
        )
        long_term_gains_on_collectibles_deduction = (
            tax.capital_gains.long_term_collectibles_deduction
            * nonnegative_long_term_capital_gains_on_collectibles
        )

        deductions = (
            short_term_loss_against_interest_dividends
            + long_term_loss_against_interest_dividends
            + long_term_loss_against_short_term_gain
            + long_term_gains_on_collectibles_deduction
        )
        return max_(0, part_a_gross_income - deductions)



from policyengine_us.model_api import *


class il_liheap_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Illinois LIHEAP income eligible"
    definition_period = YEAR
    defined_for = StateCode.IL
    reference = "https://dceo.illinois.gov/communityservices/utilitybillassistance.html"

    def formula(spm_unit, period, parameters):
        p_hhs = parameters(period).gov.hhs
        p_il = parameters(period).gov.states.il.dceo.liheap

        # Illinois uses 60% of state median income or 200% FPL, whichever is higher
        state_median_income = spm_unit("hhs_smi", period)
        smi_limit = state_median_income * p_hhs.liheap.smi_limit

        # Federal poverty guideline limit
        fpg = spm_unit("spm_unit_fpg", period)
        fpg_limit = fpg * p_il.eligibility.fpg_limit

        # Income concept - using IRS gross income as the base
        income = add(spm_unit, period, ["irs_gross_income"])

        # Eligible if income is below the higher of the two limits
        income_limit = max_(smi_limit, fpg_limit)

        return income <= income_limit



from policyengine_us.model_api import *


class mo_pension_and_ss_or_ssd_deduction_section_a(Variable):
    value_type = float
    entity = Person
    label = "Missouri Pension and Social Security or SS Disability Deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf#page=3",
        "https://dor.mo.gov/forms/MO-1040%20Fillable%20Calculating_2021.pdf#page=2",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.124",
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        # NOTE: Excludes military pensions, per: https://dor.mo.gov/forms/Military%20Reference%20Guide.pdf#page=11
        mo_agi = person("mo_adjusted_gross_income", period)
        tax_unit = person.tax_unit
        tax_unit_mo_agi = tax_unit.sum(mo_agi)
        taxable_social_security_benefits = person(
            "taxable_social_security", period
        )
        tax_unit_taxable_social_security_benefits = tax_unit.sum(
            taxable_social_security_benefits
        )
        agi_in_excess_of_taxable_social_security = (
            tax_unit_mo_agi - tax_unit_taxable_social_security_benefits
        )  # Equivalent to Line 3 of section A and B
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.mo.tax.income.deductions.social_security_and_public_pension

        # Section A, Public Pension Amounts
        # TODO:
        # unclear reference to "See instructions if Line 3 of Section C is more than $0" here: https://dor.mo.gov/forms/MO-A_2021.pdf#page=3
        public_pension_allowance = p.mo_public_pension_deduction_allowance[
            filing_status
        ]
        agi_over_public_pension__allowance = max_(
            agi_in_excess_of_taxable_social_security
            - public_pension_allowance,
            0,
        )
        public_pension_amount = person("taxable_public_pension_income", period)
        max_social_security_benefit = (
            p.mo_max_social_security_benefit
        )  # Seen on Line 7, Section A
        public_pension_value = min_(
            public_pension_amount, max_social_security_benefit
        )
        ss_or_ssdi_exemption_threshold = p.mo_ss_or_ssdi_exemption_threshold[
            filing_status
        ]

        eligible_ss_or_ssd = person(
            "mo_pension_and_ss_or_ssd_deduction_section_c", period
        )

        # From instructions here: https://dor.mo.gov/forms/MO-1040%20Instructions_2021.pdf#page=17
        if p.income_threshold_applies:
            adjusted_ss_or_ssdi_value = where(
                (mo_agi - ss_or_ssdi_exemption_threshold) > 0,
                eligible_ss_or_ssd,
                # this comes from the result of section c
                taxable_social_security_benefits,  # this is the unmodified benefits value, from Part 3 - Section C, Line 6
            )
            public_pension_less_ss_deduction = max_(
                public_pension_value - adjusted_ss_or_ssdi_value, 0
            )

            return max_(
                public_pension_less_ss_deduction
                - agi_over_public_pension__allowance,
                0,
            )
        return max_(public_pension_value - eligible_ss_or_ssd, 0)



from policyengine_us.model_api import *


class dc_gac_eligible_child(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible child for DC General Assistance for Children (GAC) based on demographics"
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.05a"
    )
    defined_for = "dc_pap_eligible_child"

    def formula(person, period, parameters):
        return ~person("is_related_to_head_or_spouse", period)



from policyengine_us.model_api import *


class ut_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah income tax before non-refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = (
        "https://tax.utah.gov/forms/current/tc-40.pdf#page=1"  # line 22
    )

    def formula(tax_unit, period, parameters):
        taxpayer_credit = tax_unit("ut_taxpayer_credit", period)
        ut_income_tax_before_credits = tax_unit(
            "ut_income_tax_before_credits", period
        )
        is_tax_liable = ~tax_unit("ut_income_tax_exempt", period)
        net_income_tax_before_credits = max_(
            ut_income_tax_before_credits - taxpayer_credit, 0
        )
        return is_tax_liable * net_income_tax_before_credits



from policyengine_us.model_api import *


class ia_fedtax_deduction(Variable):
    value_type = float
    entity = Person
    label = "Iowa deduction for selected components of federal income tax"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=41"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=41"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        # Iowa allows a deduction of federal income taxes from Iowa net
        # income, but federal payroll taxes (FICA and SECA and Additional
        # Medicare) cannot be used to reduced Iowa net income.
        ustax = person.tax_unit("income_tax_before_refundable_credits", period)
        # correct ustax amount for flaw in CTC refundable/nonrefundable split
        ctc = parameters(period).gov.irs.credits.ctc
        ctc_fully_refundable = ctc.refundable.fully_refundable
        if ctc_fully_refundable:
            us_tax = ustax + person.tax_unit("non_refundable_ctc", period)
        else:
            us_tax = ustax
        # remove SECA and Additional Medicare taxes
        amtax = person.tax_unit("additional_medicare_tax", period)
        setax = add(person.tax_unit, period, ["self_employment_tax"])
        agg_tax = max_(0.0, us_tax - amtax - setax)
        # project agg_tax to head and spouse in net_tax variable
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        net_tax = (is_head | is_spouse) * agg_tax
        # prorate net_tax among head and spouse according to net incomes
        fraction = person("ia_prorate_fraction", period)
        return fraction * net_tax



from policyengine_us.model_api import *


class ca_calworks_child_care_weeks_per_month(Variable):
    value_type = int
    entity = Person
    label = "California CalWORKs Child Care weeks per month"
    definition_period = MONTH
    defined_for = StateCode.CA



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_ctc_expansion() -> Reform:
    class refundable_ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "refundable CTC"
        unit = USD
        documentation = (
            "The portion of the Child Tax Credit that is refundable."
        )
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/24#d"

        def formula(tax_unit, period, parameters):
            # This line corresponds to "the credit which would be allowed under this section [the CTC section]"
            # without regard to this subsection [the refundability section] and the limitation under
            # section 26(a) [the section that limits the amount of the non-refundable CTC to tax liability].
            # This is the full CTC. This is then limited to the maximum refundable amount per child as per the
            # TCJA provision.

            p_ctc = parameters(period).gov.irs.credits.ctc
            p_wyden_smith = parameters(period).gov.contrib.congress.wyden_smith

            maximum_amount = tax_unit("ctc_refundable_maximum", period)

            if p_ctc.refundable.fully_refundable:
                reduction = tax_unit("ctc_phase_out", period)
                return max_(0, maximum_amount - reduction)

            total_ctc = tax_unit("ctc", period)
            maximum_refundable_ctc = min_(maximum_amount, total_ctc)

            # The other part of the "lesser of" statement is: "the amount by which [the non-refundable CTC]
            # would increase if [tax liability] increased by tax_increase", where tax_increase is the greater of:
            # - the phase-in amount
            # - Social Security tax minus the EITC
            # First, we find tax_increase:

            current_year_earnings = tax_unit("tax_unit_earned_income", period)
            if p_wyden_smith.actc_lookback:
                prior_year_earnings = tax_unit(
                    "tax_unit_earned_income_last_year", period
                )
                earnings = max_(current_year_earnings, prior_year_earnings)
            else:
                earnings = current_year_earnings
            earnings_over_threshold = max_(
                0, earnings - p_ctc.refundable.phase_in.threshold
            )
            qualifying_children = tax_unit("ctc_qualifying_children", period)

            if p_wyden_smith.per_child_actc_phase_in:
                phase_in_rate = (
                    p_ctc.refundable.phase_in.rate * qualifying_children
                )
            else:
                phase_in_rate = p_ctc.refundable.phase_in.rate

            relevant_earnings = earnings_over_threshold * phase_in_rate

            # Compute "Social Security taxes" as defined in the US Code for the ACTC.
            # This includes OASDI and Medicare payroll taxes, as well as half
            # of self-employment taxes.
            SS_ADD_VARIABLES = [
                # Person:
                "employee_social_security_tax",
                "employee_medicare_tax",
                "unreported_payroll_tax",
                # Tax unit:
                "self_employment_tax_ald",
                "additional_medicare_tax",
            ]
            SS_SUBTRACT_VARIABLES = ["excess_payroll_tax_withheld"]
            social_security_tax = add(
                tax_unit, period, SS_ADD_VARIABLES
            ) - add(tax_unit, period, SS_SUBTRACT_VARIABLES)
            eitc = tax_unit("eitc", period)
            social_security_excess = max_(0, social_security_tax - eitc)
            qualifying_children = tax_unit("ctc_qualifying_children", period)
            tax_increase = where(
                qualifying_children
                < p_ctc.refundable.phase_in.min_children_for_ss_taxes_minus_eitc,
                relevant_earnings,
                max_(relevant_earnings, social_security_excess),
            )
            limiting_tax = tax_unit("ctc_limiting_tax_liability", period)
            ctc_capped_by_tax = min_(total_ctc, limiting_tax)
            ctc_capped_by_increased_tax = min_(
                total_ctc, limiting_tax + tax_increase
            )
            amount_ctc_would_increase = (
                ctc_capped_by_increased_tax - ctc_capped_by_tax
            )

            return min_(maximum_refundable_ctc, amount_ctc_would_increase)

    class reform(Reform):
        def apply(self):
            self.update_variable(refundable_ctc)

    return reform


def create_ctc_expansion_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_ctc_expansion()

    # Look ahead for the next five years

    p = parameters.gov.contrib.congress.wyden_smith

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if (
            p(current_period).actc_lookback
            or p(current_period).per_child_actc_phase_in
        ):
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_ctc_expansion()
    else:
        return None


ctc_expansion = create_ctc_expansion_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class nj_property_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey property tax deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-3a-17/"
    defined_for = "nj_taking_property_tax_deduction"

    def formula(tax_unit, period, parameters):
        deduction = tax_unit("nj_potential_property_tax_deduction", period)
        taking_deduction = tax_unit("nj_taking_property_tax_deduction", period)
        return deduction * taking_deduction



from policyengine_us.model_api import *


class meets_school_meal_categorical_eligibility(Variable):
    value_type = bool
    entity = SPMUnit
    label = "School meal categorical eligibility"
    documentation = "Whether this SPM unit is eligible for free school meal via participation in other programs"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/7/245.2"
    adds = "gov.usda.school_meals.categorical_eligibility"



from policyengine_us.model_api import *


class hours_worked_last_week(Variable):
    value_type = float
    entity = Person
    label = "weekly hours worked on the previous week"
    unit = "hour"
    definition_period = YEAR



from policyengine_us.model_api import *


class mt_income_tax_joint(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana income tax when married couples are filing jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = ["mt_income_tax_before_refundable_credits_joint"]
    subtracts = ["mt_refundable_credits"]



from policyengine_us.model_api import *


class in_county_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana county tax"
    definition_period = YEAR
    unit = USD
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-2-1"  # (a)(3)
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        # County calculations are at the person level for each taxpayer in the law
        in_in = tax_unit.household("state_code_str", period) == "IN"
        county = tax_unit.household("county_str", period)
        rate = np.zeros_like(county, dtype=float)
        rates = parameters(period).gov.states["in"].tax.income.county_rates
        rate[in_in] = rates[county[in_in]]
        return rate * tax_unit("in_agi", period)



from policyengine_us.model_api import *


class ca_calworks_child_care_child_age_eligible(Variable):
    value_type = bool
    entity = Person
    label = (
        "Eligible child for the California CalWORKs Child Care based on age"
    )
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_Overview%2F1210_Overview.htm%23Backgroundbc-3&rhtocid=_3_3_0_2"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ca.cdss.tanf.child_care.eligibility
        age = person("age", period)
        is_disabled = person("is_disabled", period)
        age_limit = where(
            is_disabled, p.disabled_age_threshold, p.age_threshold
        )
        return age <= age_limit



from policyengine_us.model_api import *


class nm_cdcc_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Eligible child for the New Mexico dependent child day care credit"
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503752/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgDsfAEwBKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = StateCode.NM

    def formula(person, period, parameters):
        p = parameters(period).gov.states.nm.tax.income.credits.cdcc
        age = person("age", period)
        age_eligible = age < p.age_eligible
        dependent = person("is_tax_unit_dependent", period)
        return age_eligible & dependent



from policyengine_us.model_api import *


class tax_unit_child_dependents(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of child dependents in the tax unit"
    definition_period = YEAR

    adds = ["is_child_dependent"]



from policyengine_us.model_api import *


class claimed_as_dependent_on_another_return(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Is claimed as a dependent elsewhere"
    documentation = (
        "Whether the person is claimed as a dependent in another tax unit."
    )



from policyengine_core.model_api import *
from policyengine_core.periods import instant


def modify_parameters(parameters):
    parameters.gov.usda.snap.emergency_allotment.in_effect.CA.update(
        start=instant("2022-04-08"), stop=instant("2027-12-31"), value=0
    )
    return parameters


class abolish_ca_snap_ea_may_22(Reform):
    def apply(self):
        self.modify_parameters(modify_parameters)



from policyengine_us.model_api import *


class nh_total_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Hampshire total exemption allowance"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NH

    adds = [
        "nh_base_exemption",
        "nh_blind_exemption",
        "nh_disabled_exemption",
        "nh_old_age_exemption",
    ]



from policyengine_us.model_api import *


class ut_personal_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah personal exemption"
    unit = USD
    defined_for = StateCode.UT
    definition_period = YEAR
    reference = (
        "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S114.html?v=C59-10-S114_2022032320220323",  # Form TC-40, Line 11
        "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1018.html",  # 59-10-1018 (1)(g)
        "https://tax.utah.gov/forms/current/tc-40inst.pdf#page=4",  # What's New, Additional dependent for taxpayer tax credit
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ut.tax.income.credits.taxpayer
        total_dependents = tax_unit("ut_total_dependents", period)
        if p.in_effect:
            additional_dependents = tax_unit(
                "ut_personal_exemption_additional_dependents", period
            )
            total_dependents += additional_dependents
        return p.personal_exemption * total_dependents



from policyengine_us.model_api import *


class pa_income_tax_after_forgiveness(Variable):
    value_type = float
    entity = TaxUnit
    label = "PA income tax after forgiveness"
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2021/2021_pa-40in.pdf#page=21"
    defined_for = StateCode.PA

    adds = ["pa_income_tax_before_forgiveness"]
    subtracts = ["pa_tax_forgiveness_amount"]



from policyengine_us.model_api import *


class va_eitc_person(Variable):
    value_type = float
    entity = Person
    label = "Virginia Earned Income Tax Credit per individual when married filing seperately"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=32"
    defined_for = StateCode.VA

    def formula(person, period, parameters):
        refundable_eitc = person.tax_unit("va_refundable_eitc", period)
        non_refundable_eitc = person.tax_unit("va_non_refundable_eitc", period)
        claims_refundable = person.tax_unit(
            "va_claims_refundable_eitc", period
        )
        va_agi_share = person("va_agi_share", period)
        eitc_claimed = where(
            claims_refundable, refundable_eitc, non_refundable_eitc
        )
        return eitc_claimed * va_agi_share



from policyengine_us.model_api import *


class ca_is_qualifying_child_for_caleitc(Variable):
    value_type = bool
    entity = Person
    label = "Child qualifies for CalEITC"
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/file/personal/credits/EITC-calculator/Help/QualifyingChildren"
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        # CalEITC uses federal EITC rules regarding qualifying children
        return person("is_child_dependent", period)



from policyengine_us.model_api import *


class nc_child_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina child deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        # calculate deduction amount per eligible child
        federal_agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        p = parameters(period).gov.states.nc.tax.income.deductions.child
        amount = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.JOINT,
                filing_status == statuses.SURVIVING_SPOUSE,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.single.calc(federal_agi, right=True),
                p.separate.calc(federal_agi, right=True),
                p.joint.calc(federal_agi, right=True),
                p.surviving_spouse.calc(federal_agi, right=True),
                p.head_of_household.calc(federal_agi, right=True),
            ],
        )
        # calculate number of eligible children
        children = tax_unit("ctc_qualifying_children", period)
        # calculate NC child deduction
        return amount * children



from policyengine_us.model_api import *


class student_loan_interest(Variable):
    value_type = float
    entity = Person
    label = "Student loan interest expense"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class phone_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "Phone expense"
    unit = USD
    definition_period = YEAR

    adds = ["phone_cost"]  # For compatibility



from policyengine_us.model_api import *


class or_income_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR income subtractions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=13",
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # Subsection 316.800
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["or"].tax.income.subtractions
        total_subtractions = add(tax_unit, period, p.subtractions)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class pre_subsidy_transportation_expense(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Pre-subsidy transportation expense"
    unit = USD



from pathlib import Path
import pandas as pd
import os

ZIP_CODE_DATASET_PATH = Path(__file__).parent / "zip_codes.csv.gz"

# Avoid circular import error when -us-data is initialized
if os.path.exists(ZIP_CODE_DATASET_PATH):
    ZIP_CODE_DATASET = pd.read_csv(ZIP_CODE_DATASET_PATH, compression="gzip")
else:
    ZIP_CODE_DATASET = None
